from typing import Type, TypeVar

import fsspec
import numpy as np
import torch
from tensordict import TensorDictBase
from tensordict.utils import DeviceType
from torch.utils.data import Dataset

from causica.datasets.causica_dataset_format.load import (
    Variable,
    VariablesMetadata,
    get_categorical_sizes,
    tensordict_from_variables_metadata,
)
from causica.datasets.tensordict_utils import convert_one_hot


def load_adjacency_matrix(path: str) -> torch.Tensor:
    """Load an adjacency matrix from a file path.

    Args:
        path: Path to a .csv or .npy file

    Returns:
        Loaded adjacency matrix.
    """
    if path.endswith(".csv"):
        with fsspec.open_files(path, "r") as files:
            (file,) = files
            return torch.tensor(np.loadtxt(file, dtype=int, delimiter=","))
    elif path.endswith(".npy"):
        with fsspec.open_files(path, "rb") as files:
            (file,) = files
            return torch.tensor(np.load(file))
    else:
        raise ValueError("Unsupported file format.")


def ensure_adjacency_matrix(
    adjacency_matrix: str | np.ndarray | torch.Tensor,
) -> torch.Tensor:
    """Ensure that the adjacency matrix is a tensor.

    Args:
        adjacency_matrix: Adjacency matrix or path to CSV file with adjacency matrix.

    Returns:
        Adjacency matrix as a tensor.
    """
    if isinstance(adjacency_matrix, str):
        adjacency_matrix = load_adjacency_matrix(adjacency_matrix)
    elif isinstance(adjacency_matrix, np.ndarray):
        adjacency_matrix = torch.tensor(adjacency_matrix)
    return adjacency_matrix


def ensure_variables_metadata(
    variables_metadata: str | VariablesMetadata,
) -> VariablesMetadata:
    """Ensure that the variables metadata is loaded.

    Args:
        variables_metadata: Variables metadata or path to JSON file with variables metadata.

    Returns:
        Variables metadata.
    """
    if isinstance(variables_metadata, str):
        with fsspec.open(variables_metadata, "r") as f:
            return VariablesMetadata.from_json(f.read())  # type: ignore
    return variables_metadata


def preprocess_data(
    data: str | np.ndarray | torch.Tensor | TensorDictBase, variables_metadata: VariablesMetadata | None = None
) -> TensorDictBase:
    """Preprocess and if necessary load and format the data.

    Args:
        data: Data or path to a CSV file with data and no header.
        variables_metadata: Variables metadata, if provided will inform the names and preprocessing of variables and
            only select the ones present here. Otherwise all will be assumed continuous variables and named by their
            index.

    Returns:
        Data is formatted as if read by `tensordict_from_variables_metadata`.
    """
    # First, collapse types so that data is torch.Tensor | TensorDictBase
    if isinstance(data, str):
        with fsspec.open(data, "r") as f:
            data = np.loadtxt(f, delimiter=",")
    if isinstance(data, np.ndarray):
        data = torch.from_numpy(data)

    # Then, ensure data is a TensorDict with variables following the metadata
    if isinstance(data, torch.Tensor):
        if variables_metadata is None:
            variables_metadata = VariablesMetadata([Variable(f"x{i}", f"x{i}") for i in range(data.shape[1])])
        data = tensordict_from_variables_metadata(data, variables_metadata.variables)
    elif variables_metadata is not None:
        data = data.select(*(variable.name for variable in variables_metadata.variables))
    else:  # No metadata and data is already a TensorDict
        return data

    # One-hot encode categorical variables if indicated by variables metadata
    categorical_sizes = get_categorical_sizes(variables_list=variables_metadata.variables)
    return convert_one_hot(data, one_hot_sizes=categorical_sizes)


TD = TypeVar("TD", bound=TensorDictBase)


def index_contiguous_chunks(td: TD, index_key: str) -> tuple[TD, dict[int, slice]]:
    """Index a TensorDict into contiguous chunks maintaing the relative order by one of its values.

    Args:
        td: TensorDict to index.
        index_key: Key to the index value in td. The value must be scalar or 1D apart from its batch shape.

    Example:
        >> td = TensorDict({"index": torch.tensor([0, 0, 1, 1, 0, 2, 2]), "x": torch.tensor([4, 5, 6, 7, 8, 9, 10])})
        >> index_contiguous_chunks(td, "index")
        ({x: torch.tensor([4, 5, 8, 6, 7, 9, 10])}, {0: slice(0, 3), 1: slice(3, 5), 2: slice(5, 7)})

    Returns:
        A contigous copy of the TensorDict without the index value.
        A mapping from index to slice which can be used to directly map to the chunk in the TensorDict.
    """
    # Get the series index and sort to maintain order within series but make sure that all series are contiguous
    td = td.clone(recurse=False)
    index = td.pop(index_key)
    if index.dim() > 2 or (index.dim() == 2 and index.shape[1] > 1):
        raise ValueError("Pivot index must be scalar or only have one dimension across the batch.")
    if index.dim() == 2:
        index = index.squeeze(1)

    # Order by the index and build contiguous copies
    index, order = torch.sort(index, stable=True)
    td = td[order].contiguous()

    # Compute timeseries locations
    series_boundaries = torch.diff(index, dim=0)
    series_locs, *_ = torch.nonzero(series_boundaries, as_tuple=True)
    # Add the first and last index to build the ranges
    series_ranges = torch.cat(
        [
            torch.zeros((1,), dtype=series_locs.dtype),
            series_locs + 1,
            torch.tensor([len(index)], dtype=series_locs.dtype),
        ]
    )

    # Build a mapping to lookup indices quickly
    return td, {
        index: slice(a.item(), b.item()) for index, (a, b) in enumerate(zip(series_ranges[:-1], series_ranges[1:]))
    }


CLS = TypeVar("CLS", bound="IndexedTimeseriesDataset")


class IndexedTimeseriesDataset(Dataset):
    """Represents a timeseries dataset where each series is indexed by a unique identifier and the steps given by order.

    The raw dataset can be provided in the following formats:
    - TensorDict: the expected batch size is the total number of steps from all series, and the key / values refer
        to individual features.
    - Tensor or filepaths: a tensor type or a path to a CSV file with a table. Both with the expected shape
        [`total number of steps from all series`, `num features`].

    The relative order of each series is maintened and the provided dataset must therefore already be in timestep order.
    """

    @torch.no_grad()
    def __init__(
        self,
        series_index_key: str | int,
        data: str | np.ndarray | torch.Tensor | TensorDictBase,
        adjacency_matrix: str | np.ndarray | torch.Tensor,
        variables_metadata: str | VariablesMetadata | None = None,
        device: None | DeviceType = None,
    ) -> None:
        """
        Args:
            series_index_key: Name of the variable that contains the series index or the index to the key / column.
            data: Data or path to the CSV file with data.
            adjacency_matrix: Adjacency matrix or path to CSV file with adjacency matrix.
            variables_metadata: Variables metadata or path to JSON file with variables metadata. If not provided and
                variable names cannot be inferred from the data, the variables will be named x0, x1, ... in order and
                interpreted as continuous variables.
            device: Device to move the data to.
        """
        super().__init__()

        # Load and fix types of inputs
        if variables_metadata is not None:
            variables_metadata = ensure_variables_metadata(variables_metadata)
        adjacency_matrix = ensure_adjacency_matrix(adjacency_matrix)
        data = preprocess_data(data, variables_metadata)
        if isinstance(series_index_key, int):
            series_index_key = list(data.keys())[series_index_key]
            assert isinstance(series_index_key, str)

        data[series_index_key] = data[series_index_key].to(dtype=torch.int64)
        if device is not None:
            data = data.to(device)
            adjacency_matrix = adjacency_matrix.to(device=device)

        self._data, self._series_lookup = index_contiguous_chunks(data, series_index_key)
        self._adjacency_matrix = adjacency_matrix

    @classmethod
    def from_dense(cls: Type[CLS], data: torch.Tensor, lengths: torch.Tensor, **kwargs) -> CLS:
        """Create an indexed timeseries dataset from a dense representation of data with variable lengths.

        With this representation each timeseries i contains the data of data[i, lengths[i]].

        Args:
            data: Dense timeseries data of shape (num_timeseries, max_length, num_dims).
            lengths: Length of each timeseries in the data, shape (num_timeseries, ).
            **kwargs: Forwarded to `__init__`.

        Returns:
            IndexedTimeseriesDataset with the given data.
        """
        if data.dim() != 3:
            raise ValueError(f"The expected shape of data is (num_timeseries, max_length, num_dims), got {data.shape}")
        if lengths.dim() != 1:
            raise ValueError(f"The expected shape of lengths is (num_timeseries, ), got {lengths.shape}")
        if data.shape[0] != lengths.shape[0]:
            raise ValueError(
                f"The number of timeseries in data ({data.shape[0]}) does not match the number of entries "
                f"in lengths ({lengths.shape[0]})"
            )

        num_timeseries, max_length, num_dims = data.shape
        index = torch.arange(num_timeseries)[:, None].expand(num_timeseries, max_length)

        # Create sequence mask of shape (num_timeseries, max_length, num_dims) which select sequences from start to end,
        # and where all entries have the same value along the last axis.
        mask = torch.arange(max_length)[None, :, None].expand(1, max_length, num_dims) < lengths[:, None, None]

        # Use mask to create sparse data
        sparse_data = torch.cat([data[mask].reshape(-1, num_dims), index[mask[:, :, 0]].reshape(-1, 1)], dim=-1)
        return cls(series_index_key=num_dims, data=sparse_data, **kwargs)

    def __getitem__(self, idx) -> TensorDictBase:
        """Get timeseries by their relative index.

        Args:
            idx: Relative index in [0, len(self)]

        Returns:
            The full timeseries TensorDict(..., batch_size=`number of steps`).
        """
        return self._data[self._series_lookup[idx]]

    def __len__(self) -> int:
        """The total number of timeseries in the dataset."""
        return len(self._series_lookup)



import abc
from typing import Iterable, Optional, Sequence

import torch
import torch.distributions as dist
from tensordict import TensorDict


class SEM(dist.Distribution, abc.ABC):
    """A structural equation model (SEM).

    A SEM defines the causal relationships amongst a given set of nodes. This class provides methods to sample data
    from the observational and interventional distributions of the SEM.

    A SEMs have a batch shape: `batch_shape_f + batch_shape_g`

    In general, a SEM has a batch shape corresponding to `functions shape + graphs shape` specified as `batch_shape_f`
    and `batch_shape_g` respectively. This allows for batched processing of different functions (e.g. interventions on
    the same nodes) on different graphs. Samples and noise are broadcast to include this shape before calling the
    functional relationships.
    """

    arg_constraints: dict = {}

    def __init__(
        self,
        graph: torch.Tensor,
        node_names: Sequence[str],
        event_shape: torch.Size = torch.Size(),
        batch_shape: torch.Size | None = None,
    ) -> None:
        """
        Args:
            graph: Adjacency Matrix for the underlying SEM
            node_names: The names of each node in the graph
            event_shape: Shape of a sample from the distribution. This is currently not used because the output shape is
                determined by the dictionary holding the samples.
            batch_shape: Batch shape of the distribution. This decomposes into the batch shape of the functions and the
                batch shape of the graphs: batch_shape = batch_shape_f + batch_shape_g. If None, the batch shape of the
                graph is used.
        """
        self._graph = graph
        self.node_names = node_names
        assert node_names, "Can't have an empty SEM"
        assert graph.shape[-2:] == (
            len(node_names),
            len(node_names),
        ), "Graph adjacency matrix must have shape [num_nodes, num_nodes] (excluding batch dimensions)"
        if batch_shape is None:
            self.batch_shape_g = torch.Size(graph.shape[:1]) if graph.ndim > 2 else torch.Size([])
            batch_shape = self.batch_shape_g
            self.batch_shape_f = torch.Size([])
        super().__init__(event_shape=event_shape, batch_shape=batch_shape)

    @property
    def graph(self) -> torch.Tensor:
        """
        The adjacency matrix representing the graph.
        """
        return self._graph

    @abc.abstractmethod
    def log_prob(self, value: TensorDict) -> torch.Tensor:
        pass

    @abc.abstractmethod
    def do(self, interventions: TensorDict) -> "SEM":
        """Return the SEM associated with the interventions.

        If the interventions have a batch shape, then the returned functions will have the same batch shape to batch
        the computation across interventions.
        """

    def condition(self, conditioning: TensorDict) -> "SEM":
        """Return a SEM with conditioned variables.

        This implementation of conditioning relies on the do operator and only supports conditioning on a conditioning
        set with no unconditioned parents. This is a limitation of the current implementation and will be lifted in the
        future. For example in a SEM with the following graph: A -> B -> C and A -> C, conditioning on B is not
        supported yet. However, conditioning on A or {A, B} is supported.

        Args:
            conditioning: A dictionary of conditioning values for the nodes in the SEM. There must not be any nodes
                outside of the conditioning set that are parents of the conditioning set.

        Return:
            The SEM conditioned on the conditioning set
        """
        conditioning_node_names = set(conditioning.keys())
        unconditioned_node_names = [
            name for name in self.node_names if name not in conditioning_node_names
        ]  # keep order

        unconditioned_idx = [self.node_names.index(name) for name in unconditioned_node_names]

        for i, name in enumerate(self.node_names):
            if name in conditioning_node_names and self.graph[unconditioned_idx, i].sum() > 0:
                parents = set(node for idx, node in enumerate(self.node_names) if self.graph[idx, i] > 0)
                unconditioned_parents = parents - conditioning_node_names
                raise NotImplementedError(
                    f"Conditioning on a node with unconditioned parents is not supported yet. Tried conditioning on {name} with parents {unconditioned_parents}"
                )

        return self.do(conditioning)

    @abc.abstractmethod
    def noise_to_sample(self, noise: TensorDict) -> TensorDict:
        """
        For a given noise vector, return the corresponding sample values.

        Args:
            noise: Dictionary of tensors representing the noise shape sample_shape + batch_shape + noise_shape
        Return:
            Dictionary of samples from the sem corresponding to the noise, shape sample_shape + batch_shape + event_shape
        """

    @abc.abstractmethod
    def sample_to_noise(self, sample: TensorDict) -> TensorDict:
        """
        For a given sample get the noise vector.

        Args:
            sample: samples from the sem each of shape sample_shape + batch_shape + event_shape
        Return:
            Dictionary of tensors representing the noise shape sample_shape + batch_shape + noise_shape
        """

    @torch.no_grad()
    def sample(self, sample_shape: torch.Size = torch.Size()) -> TensorDict:
        """
        Sample from the SEM

        Grads shall not pass through this method (see `Distribution.sample` vs `Distribution.rsample`).

        Args:
            sample_shape: shape of the returned samples
        Return:
            Dictionary of samples from the sem corresponding to the noise, shape sample_shape + batch_shape + event_shape
        """
        noise_dict = self.sample_noise(sample_shape=sample_shape)
        return self.noise_to_sample(noise=noise_dict)

    @abc.abstractmethod
    @torch.no_grad()
    def sample_noise(self, sample_shape: torch.Size = torch.Size()) -> TensorDict:
        """
        Sample the noise vector for the distribution.

        Grads shall not pass through implementations of this method.

        Args:
            sample_shape: shape of the returned noise samples
        Return:
            Dictionary of samples from the noise distribution for each node, shape sample_shape + batch_shape + event_shape
        """


def ite(
    sem: SEM,
    factual_data: TensorDict,
    intervention_a: TensorDict,
    intervention_b: TensorDict,
    effects: Optional[Iterable[str]] = None,
) -> TensorDict:
    """Calculate ITE of intervention A and B on some factual data for a list of effects.

    Args:
        factual_data: Factual data to abduct the noise from.
        intervention_a: Specification of intervention A.
        intervention_b: Specification of intervention B.
        effects: List of effect variables. Defaults to None.

    Returns:
       TensorDict: Dictionary holding the ITEs for the effect variables for all samples.
    """
    if effects is None:
        effects = list(set(sem.node_names) - set(intervention_a.keys()) - set(intervention_b.keys()))

    base_noise = sem.sample_to_noise(factual_data)
    do_a_cfs = sem.do(interventions=intervention_a).noise_to_sample(base_noise)
    do_b_cfs = sem.do(interventions=intervention_b).noise_to_sample(base_noise)

    return {effect: do_a_cfs[effect] - do_b_cfs[effect] for effect in effects}


def counterfactual(sem: SEM, factual_data: TensorDict, intervention: TensorDict) -> TensorDict:
    """Calculate Counterfactual of an intervention on some factual data.

    Args:
        factual_data: Factual data to abduct the noise from.
        intervention: Specification of intervention.

    Returns:
       TensorDict: Dictionary holding the counterfactual values for all samples.
    """
    return sem.do(interventions=intervention).noise_to_sample(sem.sample_to_noise(factual_data))


def ate(
    sem: SEM,
    intervention_a: TensorDict,
    intervention_b: TensorDict,
    effects: Optional[Iterable[str]] = None,
    num_samples: int = 1000,
) -> TensorDict:
    """Calculate the ATE of intervention A and B for a list of effects.

    Args:
        intervention_a: Specification of intervention A.
        intervention_b: Specification of intervention B.
        effects: List of effect variables. Defaults to None.
        num_samples: Number of Monte-Carlo samples to estimate the ATE. Defaults to 1000.

    Returns:
       TensorDict: Dictionary holding the ATEs for the effect variables.
    """
    sample_shape = torch.Size([num_samples])

    if effects is None:
        effects = list(set(sem.node_names) - set(intervention_a.keys()) - set(intervention_b.keys()))

    do_a_samples = sem.do(interventions=intervention_a).sample(sample_shape)
    do_b_samples = sem.do(interventions=intervention_b).sample(sample_shape)

    return TensorDict(
        {effect: do_a_samples[effect].mean(0) - do_b_samples[effect].mean(0) for effect in effects}, batch_size=tuple()
    )



"""Module that provides data normalization functionality."""
from typing import Any, Callable, Optional

import torch
import torch.distributions as td
from tensordict import TensorDictBase
from torch import nn
from torch.distributions import constraints

from causica.distributions.transforms import JointTransformModule, SequentialTransformModule, TransformModule

Normalizer = TransformModule[TensorDictBase, TensorDictBase]
FitNormalizerType = Callable[[TensorDictBase], Normalizer]


class LoadNoneTensorMixin(nn.Module):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._register_load_state_dict_pre_hook(self._update_tensor_size_on_load)

    def _update_tensor_size_on_load(self, state_dict: dict[str, torch.Tensor], prefix: str, *args, **kwargs) -> None:
        _ = args, kwargs
        for key, value in state_dict.items():
            local_key = key.removeprefix(prefix)
            if hasattr(self, local_key) and getattr(self, local_key) is None:
                setattr(self, local_key, torch.empty_like(value))


class LogTransform(TransformModule[torch.Tensor, torch.Tensor], td.Transform, LoadNoneTensorMixin):

    """
    A transform to apply the log function to a single tensor plus an offset.
    """

    bijective = True
    domain = constraints.greater_than_eq(0)
    codomain = constraints.real

    arg_constraints = {"offset": constraints.greater_than_eq(0)}

    def __init__(self, offset: Optional[torch.Tensor]) -> None:
        """
        Args:
            offset: the offset added to the single tensor
        """
        super().__init__()
        self.offset: torch.Tensor
        self.register_buffer("offset", offset)

    def _call(self, x: torch.Tensor) -> torch.Tensor:
        return torch.log(x + self.offset)

    def _inverse(self, y: torch.Tensor) -> torch.Tensor:
        return torch.exp(y) - self.offset

    def log_abs_det_jacobian(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.log(torch.abs(1.0 / (x + self.offset)))


class Standardizer(TransformModule[torch.Tensor, torch.Tensor], td.AffineTransform, LoadNoneTensorMixin):
    """Standardizer module for a single variable, ie a single tensor."""

    def __init__(self, mean: Optional[torch.Tensor], std: Optional[torch.Tensor], *args, **kwargs) -> None:
        """
        Args:
            mean: Mean of the variable
            std: Standard deviation of the variable
            *args, **kwargs: Passed to the AffineTransform
        """
        loc = scale = None
        if mean is not None and std is not None:
            loc = -mean / std
            scale = 1 / std
        super().__init__(loc, scale, *args, **kwargs)
        del self.loc, self.scale  # Unset these temporarily to allow registering as buffers

        self.loc: torch.Tensor
        self.scale: torch.Tensor
        self.register_buffer("loc", loc)
        self.register_buffer("scale", scale)


def fit_log_normalizer(
    data: TensorDictBase, default_offset: float = 1.0, min_margin: float = 0.0, keys: Optional[list[str]] = None
) -> Normalizer:
    """Fits a log standardizer to the tensordict.

    The fitted log normalizer computes:
        log(x - min(data) * (min(data) < 0) + min + (max(data) - min(data)) * min_margin + offset).

    Args:
        data: The data to fit the standardizer to.
        default_offset: An additional offset to use. The offset is the min value + default_offset. Must be positive.
        min_margin: Adds a fraction of the range of data to the minimum offset to avoid log(x<=0) on unseen data.
        keys: Limit the set of keys to log transform if set.

    Returns:
        The log standardizer.
    """
    assert default_offset > 0, "default_offset must be positive"
    # For min_value >= 0, offset = default_offset; min_value<0 offset=abs(min_value)+default_offset
    data = data.select(*keys) if keys else data
    min_values = data.apply(lambda x: torch.min(x, dim=0, keepdim=False).values, batch_size=torch.Size())
    max_values = data.apply(lambda x: torch.max(x, dim=0, keepdim=False).values, batch_size=torch.Size())
    offsets = min_values.apply(
        lambda min_, max_: torch.where(
            min_ >= 0, default_offset * torch.ones_like(min_), torch.abs(min_) + default_offset
        )
        + (max_ - min_) * min_margin,
        max_values,
    )
    return JointTransformModule({key: LogTransform(offset) for key, offset in offsets.items()})


def fit_standardizer(data: TensorDictBase, keys: Optional[list[str]] = None) -> Normalizer:
    """Return a standardizer that updates data to zero mean and unit standard deviation."""
    data = data.select(*keys) if keys else data
    means = data.apply(
        lambda x: torch.mean(x, dim=0, keepdim=False),
        batch_size=torch.Size(),
    )
    # Filter out std == 0
    stds = data.apply(
        lambda x: torch.std(x, dim=0, keepdim=False),
        batch_size=torch.Size(),
    ).apply(lambda x: torch.where(x == 0, torch.ones_like(x), x))

    return JointTransformModule({key: Standardizer(means.get(key), stds.get(key)) for key in means.keys()})


def chain_normalizers(*fit_functions: FitNormalizerType) -> FitNormalizerType:
    """Chain a number of normalizers together.

    Args:
        *fit_functions: Functions that produce normalizers.

    Returns:
        A function that fits the sequence of normalizers.
    """

    def sequential_fitting(X: TensorDictBase) -> Normalizer:
        transform_modules = []
        for fit_function in fit_functions:
            transform_module = fit_function(X)
            X = transform_module(X)
            transform_modules.append(transform_module)
        return SequentialTransformModule[TensorDictBase, TensorDictBase](*transform_modules)

    return sequential_fitting


def infer_compatible_log_normalizer_from_checkpoint(state_dict: dict[str, Any]) -> Normalizer:
    """Infers a normalizer compatible with a model checkpoint.

    Assumes that `normalizer` is stored in the toplevel of the checkpoint and that it is a `SequentialTransformModule`,
    and may contain a `LogTransform` and a `Standardizer`. If both are set they are always in that order.

    Args:
        state_dict: The state dict of the checkpoint to load.

    Returns:
        An unitiliazed normalizer.
    """
    # Infer normalizers per variable from the checkpoint
    normalizers_dicts: dict[int, dict[str, TransformModule[torch.Tensor, torch.Tensor]]] = {0: {}, 1: {}}
    for key in state_dict.keys():
        if key.startswith("normalizer."):
            *_, variable_name, state_name = key.split(".")
            if state_name == "offset":
                normalizers_dicts[0][variable_name] = LogTransform(None)
            elif state_name in {"loc", "scale"} and variable_name not in normalizers_dicts[1]:
                normalizers_dicts[1][variable_name] = Standardizer(None, None)

    # Construct the full empty normalizer ready to be initialized from the checkpoint
    joint_transforms = [JointTransformModule(normalizer) for normalizer in normalizers_dicts.values() if normalizer]
    if not joint_transforms:
        return JointTransformModule({})  # No normalizer, return a passhtrough normalizer

    return SequentialTransformModule[TensorDictBase, TensorDictBase](*joint_transforms)



import torch
import torch.distributions as td


def signed_uniform_1d(low: float, high: float) -> td.Distribution:
    """
    Returns a distribution that is a mixture of two uniform distributions.

    The first uniform distribution is defined on the interval [low, high]
    and the second is defined on the interval [-high, -low].
    The mixture is defined by a Bernoulli distribution with a probability of 0.5.

    Args:
        low: the lower bound of the uniform distribution
        high: the upper bound of the uniform distribution

    Returns:
        a distribution that is a mixture of two uniform distributions
    """

    weight = torch.tensor([0.5, 0.5])
    mix = td.Categorical(probs=weight)
    low_tensor = torch.tensor([low, -high])
    high_tensor = torch.tensor([high, -low])
    comp = td.Uniform(low=low_tensor, high=high_tensor)
    return td.MixtureSameFamily(mix, comp)


class MultivariateSignedUniform(td.Distribution):
    """
    Distribution that is a mixture of two uniform distributions.

    The first uniform distribution is defined on the interval [low, high]
    and the second is defined on the interval [-high, -low].
    The mixture is defined by a Bernoulli distribution with a probability of 0.5.

    Args:
        low: the lower bound of the uniform distribution
        high: the upper bound of the uniform distribution
        size: the size of the distribution
    """

    def __init__(self, low: float, high: float, size: torch.Size):
        self.one_dim_dist = signed_uniform_1d(low, high)
        self.size = size
        super().__init__()

    def sample(self, sample_shape=torch.Size()):
        sample_shape = sample_shape + self.size

        return self.one_dim_dist.sample(sample_shape)



from __future__ import annotations

from typing import Optional

import torch
import torch.distributions as td

from causica.data_generation.samplers.sampler import Sampler
from causica.functional_relationships.functional_relationships import FunctionalRelationships
from causica.functional_relationships.heteroscedastic_rff_functional_relationships import (
    HeteroscedasticRFFFunctionalRelationships,
)
from causica.functional_relationships.linear_functional_relationships import LinearFunctionalRelationships
from causica.functional_relationships.rff_functional_relationships import RFFFunctionalRelationships


class FunctionalRelationshipsSampler(Sampler[FunctionalRelationships]):
    """Abstract class for sampling Functional Relationships."""

    def __init__(self, shapes_dict: dict[str, torch.Size]) -> None:
        super().__init__()

        self.shapes_dict = shapes_dict


class LinearRelationshipsSampler(FunctionalRelationshipsSampler):
    """Sample a Linear Functional Relationship, by providing a distribution for the coefficient matrix."""

    def __init__(
        self,
        scale_dist: td.Distribution,
        shapes_dict: dict[str, torch.Size],
        bias_dist: Optional[td.Distribution] = None,
    ):
        super().__init__(shapes_dict)
        self.scale_dist = scale_dist
        self.bias_dist = bias_dist

    def sample(self) -> FunctionalRelationships:
        sample_bias = self.bias_dist.sample() if self.bias_dist is not None else None
        return LinearFunctionalRelationships(
            shapes=self.shapes_dict,
            initial_linear_coefficient_matrix=self.scale_dist.sample(),
            initial_bias=sample_bias,
        )


class RFFFunctionalRelationshipsSampler(FunctionalRelationshipsSampler):
    """Sample a Non Linear Functional Relationship, by providing two distributions.

    The sampler uses two distributions:
        - a first distribution for the random features
        - a second distribution for the linear outer coefficients."""

    def __init__(
        self,
        rf_dist: td.Distribution,
        coeff_dist: td.Distribution,
        shapes_dict: dict[str, torch.Size],
        bias_dist: Optional[td.Distribution] = None,
        length_dist: Optional[td.Distribution] = None,
        out_dist: Optional[td.Distribution] = None,
        angle_dist: Optional[td.Distribution] = None,
    ):
        super().__init__(shapes_dict)
        self.bias_dist = bias_dist
        self.length_dist = length_dist
        self.out_dist = out_dist
        self.angle_dist = angle_dist
        self.rf_dist = rf_dist
        self.coeff_dist = coeff_dist

    def sample(self) -> FunctionalRelationships:
        sample_bias = self.bias_dist.sample() if self.bias_dist is not None else None
        sample_length = self.length_dist.sample() if self.length_dist is not None else None
        sample_out = self.out_dist.sample() if self.out_dist is not None else None
        sample_angle = self.angle_dist.sample() if self.angle_dist is not None else None

        return RFFFunctionalRelationships(
            shapes=self.shapes_dict,
            initial_random_features=self.rf_dist.sample(),
            initial_coefficients=self.coeff_dist.sample(),
            initial_bias=sample_bias,
            initial_length_scales=sample_length,
            initial_output_scales=sample_out,
            initial_angles=sample_angle,
        )


class HeteroscedasticRFFFunctionalRelationshipsSampler(FunctionalRelationshipsSampler):
    """Sample a Post Non Linear Functional Relationship, by providing two distributions.

    The sampler uses two distributions:
        - a first distribution for the random features
        - a second distribution for the linear outer coefficients.
    The model is of the form: x = log(1+ exp(g(x))) where g is the RFF model."""

    def __init__(
        self,
        rf_dist: td.Distribution,
        coeff_dist: td.Distribution,
        shapes_dict: dict[str, torch.Size],
        bias_dist: Optional[td.Distribution] = None,
        length_dist: Optional[td.Distribution] | Optional[float] = None,
        out_dist: Optional[td.Distribution] | Optional[float] = None,
        angle_dist: Optional[td.Distribution] = None,
        log_scale: bool = False,
    ):
        """
        Args:
            rf_dist: Distribution for the random features
            coeff_dist: Distribution for the linear outer coefficients
            shapes_dict: Shapes of the different variables
            bias_dist: Distribution for the bias
            length_dist: Distribution for the length scales
            out_dist: Distribution for the output scales
            angle_dist: Distribution for the angles
            log_scale: If True, the output is log(log(1+exp(g(x)))), else it is log(1+exp(g(x)))
        """

        super().__init__(shapes_dict)
        self.bias_dist = bias_dist
        self.length_dist = length_dist
        self.out_dist = out_dist
        self.angle_dist = angle_dist
        self.rf_dist = rf_dist
        self.coeff_dist = coeff_dist
        self.log_scale = log_scale

    def sample(self) -> HeteroscedasticRFFFunctionalRelationships:
        sample_coeff = self.coeff_dist.sample()
        sample_rf = self.rf_dist.sample()
        sample_bias = self.bias_dist.sample() if self.bias_dist is not None else None
        sample_angle = self.angle_dist.sample() if self.angle_dist is not None else None

        if self.length_dist is not None:
            if isinstance(self.length_dist, td.Distribution):
                sample_length = self.length_dist.sample()
            else:
                sample_length = self.length_dist * torch.ones_like(sample_rf[0])
        else:
            sample_length = None

        if self.out_dist is not None:
            if isinstance(self.out_dist, td.Distribution):
                sample_out = self.out_dist.sample()
            else:
                sample_out = self.out_dist * torch.ones_like(sample_rf[0])
        else:
            sample_out = None

        return HeteroscedasticRFFFunctionalRelationships(
            shapes=self.shapes_dict,
            initial_random_features=sample_rf,
            initial_coefficients=sample_coeff,
            initial_bias=sample_bias,
            initial_length_scales=sample_length,
            initial_output_scales=sample_out,
            initial_angles=sample_angle,
            log_scale=self.log_scale,
        )



import igraph as ig
import numpy as np
import torch

from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution


class EdgesPerNodeErdosRenyiDAGDistribution(AdjacencyDistribution):
    """
    An adjacency distribution for sampling Directed Acyclic Graphs using Erdos-Renyi.

    It allows to sample DAGs according to the ER distribution with a fixed number of edges per node.

    Re-implementation in pytorch from AVICI:

    https://github.com/larslorch/avici

    """

    def __init__(
        self,
        num_nodes: int,
        edges_per_node: list[int],
        validate_args: bool | None = None,
    ):
        """
        Args:
            num_nodes: the number of nodes in the DAGs to be sampled
            edges_per_node: List of the number of edges per node to sample.
        """
        assert num_nodes > 0, "Number of nodes in the graph must be greater than 0"
        super().__init__(num_nodes=num_nodes, validate_args=validate_args)
        self.edges_per_node = edges_per_node

    def _one_sample(self):
        """
        Sample a binary adjacency matrix from the underlying distribution.

        Returns:
            A tensor of shape (num_nodes, num_nodes)
        """

        get_edges_per_node = np.random.choice(self.edges_per_node).item()
        n_edges = get_edges_per_node * self.num_nodes
        prob = min(n_edges / ((self.num_nodes * (self.num_nodes - 1)) / 2), 0.99)

        mat = np.random.binomial(n=1, p=prob, size=(self.num_nodes, self.num_nodes)).astype(int)  # bernoulli

        # make DAG by zeroing above diagonal; k=-1 indicates that diagonal is zero too
        dag = np.tril(mat, k=-1)

        # randomly permute
        perm = np.random.permutation(self.num_nodes).tolist()
        dag = dag[perm, :][:, perm]

        if not ig.Graph.Weighted_Adjacency(dag.tolist()).is_dag():
            raise ValueError("Sampled graph is not a DAG")

        dag = torch.tensor(dag, dtype=torch.float32)

        return dag

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        raise NotImplementedError

    def sample(self, sample_shape: torch.Size = torch.Size()):
        """
        Sample binary adjacency matrices from the underlying distribution.

        Args:
            sample_shape: The shape of the samples to be drawn.

        Returns:
            A tensor of shape (sample_shape, num_nodes, num_nodes)
        """
        num_samples = int(torch.prod(torch.tensor(sample_shape)).item())
        return torch.stack([self._one_sample() for _ in range(num_samples)]).reshape(
            sample_shape + (self.num_nodes, self.num_nodes)
        )

    def entropy(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mean(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mode(self) -> torch.Tensor:
        raise NotImplementedError

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError



import itertools
import random
import warnings
from itertools import zip_longest
from typing import Iterable, Iterator

import torch
from torch.utils.data import BatchSampler, Sampler, SubsetRandomSampler


class SubsetBatchSampler(Sampler[list[int]]):
    """A Pytorch batch sampler that samples batches from a list of subsets.

    Each batch will be sampled from a single subset. The subsets are sampled in random order if shuffle is True.
    """

    def __init__(
        self, subset_lengths: list[int], batch_size: int, shuffle: bool = True, drop_last: bool = False
    ) -> None:
        """
        Args:
            subset_lengths: The lengths of each subset
            batch_size: The batch size to use
            shuffle: Whether to shuffle the indices
            drop_last: Whether to drop the last batch of each subset if it is smaller than the batch size.
        """
        if batch_size > min(subset_lengths):
            warnings.warn("Batch size is larger than the smallest subset length")
        if any(length < 1 for length in subset_lengths):
            raise ValueError("Subset lengths must be at least 1")

        self.subset_lengths = subset_lengths
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.start_indices = torch.cumsum(torch.tensor([0] + self.subset_lengths[:-1]), 0)

        self.batch_samplers: list[Iterator | BatchSampler]
        self.indices_lists = [
            (torch.arange(length) + offset).tolist() for length, offset in zip(self.subset_lengths, self.start_indices)
        ]
        if self.shuffle:
            self.batch_samplers = [
                BatchSampler(SubsetRandomSampler(indices), batch_size, drop_last=drop_last)
                for indices in self.indices_lists
            ]
        else:
            # Replace with `itertools.batched` when Python >= 3.12
            self.batch_samplers = []
            for indices in self.indices_lists:
                args = [iter(indices)] * batch_size
                self.batch_samplers.append(zip_longest(*args, fillvalue=None))

    def __iter__(self):
        # Support function to truncate batches to the correct size when not shuffling
        if not self.shuffle:
            yield from itertools.chain.from_iterable(
                yield_truncated_batch_from_zip(iterator, self.batch_size, self.drop_last)
                for iterator in self.batch_samplers
            )

        iterators = [iter(batch_sampler) for batch_sampler in self.batch_samplers]
        while iterators:
            permutation = list(enumerate(iterators))
            random.shuffle(permutation)
            stopped = []
            for i, iterator in permutation:
                try:
                    yield next(iterator)
                except StopIteration:
                    stopped.append(i)

            # Delete in reverse order to preserve indices
            for i in sorted(stopped, reverse=True):
                del iterators[i]

    def __len__(self) -> int:
        if self.drop_last:
            return sum(length // self.batch_size for length in self.subset_lengths)
        return sum((length + self.batch_size - 1) // self.batch_size for length in self.subset_lengths)


def yield_truncated_batch_from_zip(iterator: Iterable, batch_size: int, drop_last: bool = False):
    """Yield truncated batches from an iterator (dropping None elements), skipping the last batch if necessary."""
    for batch in iterator:
        batch = list(b for b in batch if b is not None)
        if not drop_last or len(batch) == batch_size:
            yield batch
        else:
            return



from typing import Optional

import torch
from tensordict import TensorDict

from causica.datasets.tensordict_utils import expand_td_with_batch_shape
from causica.distributions import JointNoiseModule
from causica.functional_relationships import (
    DoFunctionalRelationships,
    FunctionalRelationships,
    TemporalEmbedFunctionalRelationships,
    create_do_functional_relationship,
)
from causica.sem.structural_equation_model import SEM


class DistributionParametersSEM(SEM):
    """
    A Structural Equation Model where the functional forward pass generates the parameters of the distributions to be sampled.

    This is more general than the simple additive case.
    """

    arg_constraints: dict = {}

    def __init__(
        self,
        graph: torch.Tensor,
        noise_dist: JointNoiseModule,
        func: FunctionalRelationships,
        log_func_rescale: Optional[FunctionalRelationships] = None,
    ):
        """
        Args:
            graph: Adjacency Matrix
            noise_dist: A dictionary of callables for generating distributions from predictions
            func: The functional relationship that can map from the tensor representation to the to itself.
            log_func_rescale: The log of the functional relationship that applies a data-dependent operation on the noise.
        """

        self.batch_shape_f = func.batch_shape
        self.batch_shape_g = torch.Size(graph.shape[:-2])

        if isinstance(func, TemporalEmbedFunctionalRelationships):
            raise ValueError(f"Functional relationship {func} is a TemporalEmbedFunctionalRelationships.")

        if len(self.batch_shape_f) == 0 and len(self.batch_shape_g) > 0:
            self.batch_shape_f = torch.Size((1,))
            func.batch_shape = self.batch_shape_f

        super().__init__(
            graph=graph,
            node_names=noise_dist.keys(),
            batch_shape=torch.Size(self.batch_shape_f + self.batch_shape_g),
        )
        self.noise_dist = noise_dist
        self.func = func
        self.log_func_rescale = log_func_rescale

    def log_prob(self, value: TensorDict) -> torch.Tensor:
        """
        Compute the log prob of the observations

        Args:
            value: a dictionary of `sample_shape + batch_shape + [variable_dim]` shaped tensors
        Return:
            A tensor of shape `sample_shape + batch_shape` of the log probability of each event
        """
        expanded_value = expand_td_with_batch_shape(value, self.batch_shape)
        if self.log_func_rescale is None:
            res = self.noise_dist(self.func(expanded_value, self.graph)).log_prob(expanded_value)
        else:
            res = self.noise_dist(
                (self.func(expanded_value, self.graph), self.log_func_rescale(expanded_value, self.graph))
            ).log_prob(expanded_value)
        return res

    def noise_to_sample(self, noise: TensorDict) -> TensorDict:
        """
        For a given noise vector, return the corresponding sample values.

        Args:
            noise: Dictionary of tensors representing the noise shape
            `sample_shape + batch_shape + noise_shape`
        Return:
            Dictionary of samples from the sem corresponding to the noise, shape
            `sample_shape + batch_shape + event_shape`
        """
        x = expand_td_with_batch_shape(noise.clone(), self.batch_shape).zero_()
        expanded_noise = expand_td_with_batch_shape(noise, self.batch_shape)
        for _ in self.node_names:
            if self.log_func_rescale is None:
                x.update_(self.noise_dist(self.func(x, self.graph)).noise_to_sample(expanded_noise))
            else:
                x.update_(
                    self.noise_dist((self.func(x, self.graph), self.log_func_rescale(x, self.graph))).noise_to_sample(
                        expanded_noise
                    )
                )
        if isinstance(self.func, DoFunctionalRelationships):
            do = self.func.do
            if do.batch_dims > 0:
                do = do[(...,) + (None,) * len(self.batch_shape_g)]
            expanded_do = do.expand(*x.batch_size)
            x.update(expanded_do, inplace=True)
            x = self.func.func.tensor_to_td.order_td(x)
        return x

    def sample_to_noise(self, sample: TensorDict) -> TensorDict:
        """
        For a given sample get the noise vector.

        Args:
            sample: samples from the sem each of shape `sample_shape + batch_shape + event_shape`
        Return:
            Dictionary of tensors representing the noise shape `sample_shape + batch_shape + noise_shape`
        """
        expanded_sample = expand_td_with_batch_shape(sample, self.batch_shape)
        if self.log_func_rescale is None:
            res = self.noise_dist(self.func(expanded_sample, self.graph)).sample_to_noise(expanded_sample)
        else:
            res = self.noise_dist(
                (self.func(expanded_sample, self.graph), self.log_func_rescale(expanded_sample, self.graph))
            ).sample_to_noise(sample)

        return res

    @torch.no_grad()
    def sample_noise(self, sample_shape: torch.Size = torch.Size()) -> TensorDict:
        """
        Sample the noise vector for the distribution.

        Args:
            sample_shape: shape of the returned noise samples
        Return:
            Dictionary of samples from the noise distribution for each node, shape
            `sample_shape + batch_shape + event_shape`.
        """

        return self.noise_dist().sample(torch.Size(torch.Size(sample_shape) + self.batch_shape))

    def do(self, interventions: TensorDict) -> "DistributionParametersSEM":
        """Return the SEM associated with the interventions.

        If the interventions have a batch shape, then the returned functions will have the same batch shape to batch
        the computation across interventions.
        """
        do_log_func = None
        if isinstance(self.func, DoFunctionalRelationships):
            if self.log_func_rescale is not None:
                if not isinstance(self.log_func_rescale, DoFunctionalRelationships):
                    raise ValueError(
                        "log_func_rescale must be DoFunctionalRelationships when func is DoFunctionalRelationships"
                    )

            if set(interventions.keys()) & set(self.func.do.keys()):
                raise ValueError("Cannot intervene on already intervened nodes")
            prev_interventions = self.func.do
            if interventions.batch_dims == 0 and prev_interventions.batch_dims > 0:
                interventions = interventions.expand(prev_interventions.batch_size)
            if prev_interventions.batch_dims == 0 and interventions.batch_dims > 0:
                prev_interventions = prev_interventions.expand(interventions.batch_size)
            if interventions.batch_size != prev_interventions.batch_size:
                raise ValueError(
                    f"Interventions batch shape {interventions.batch_size} does not match do batch shape {self.func.do.batch_size}"
                )
            interventions_updated = interventions.update(prev_interventions)
            graph_updated = self.func.pad_intervened_graphs(self.graph)
            do_func, do_graphs = create_do_functional_relationship(
                interventions=interventions_updated,
                func=self.func.func,
                graph=graph_updated,
            )

            if self.log_func_rescale is not None:
                interventions_replaced = interventions_updated.apply(torch.zeros_like)
                do_log_func, _ = create_do_functional_relationship(
                    interventions=interventions_replaced,
                    func=self.log_func_rescale.func,
                    graph=graph_updated,
                )

        else:

            do_func, do_graphs = create_do_functional_relationship(
                interventions=interventions, func=self.func, graph=self.graph
            )

            if self.log_func_rescale is not None:
                interventions_replaced = interventions.apply(torch.zeros_like)
                do_log_func, _ = create_do_functional_relationship(
                    interventions=interventions_replaced, func=self.log_func_rescale, graph=self.graph
                )

        intervened_node_names = set(interventions.keys())
        unintervened_node_names = [name for name in self.node_names if name not in intervened_node_names]  # keep order
        return DistributionParametersSEM(
            graph=do_graphs,
            noise_dist=self.noise_dist[unintervened_node_names],
            func=do_func,
            log_func_rescale=do_log_func,
        )



import torch
from torch import nn


class DECIEmbedNN(nn.Module):
    """
    Defines the function f for the SEM. For each variable x_i we use
    f_i(x) = f(e_i, sum_{k in pa(i)} g(e_k, x_k)), where e_i is a learned embedding
    for node i.
    """

    def __init__(
        self,
        group_mask: torch.Tensor,
        embedding_size: int,
        out_dim_g: int,
        num_layers_g: int,
        num_layers_zeta: int,
    ):
        """
        Args:
            group_mask: A mask of shape (num_nodes, num_processed_cols) such that group_mask[i, j] = 1. when col j is in group i.
            embedding_size: Size of the embeddings used by each node. Uses the larger of 4 * concatenated_shape or embedding_size.
            out_dim_g: Output dimension of the "inner" NN, l. If none, default is embedding size.
            num_layers_g: Number of layers in the "inner" NN, l.
            num_layers_zeta: Number of layers in the "outer" NN, ζ.
        """
        super().__init__()
        self.group_mask = group_mask
        num_nodes, concatenated_shape = group_mask.shape
        # Initialize embeddings uⱼ
        self.embeddings = nn.parameter.Parameter(0.01 * torch.randn(num_nodes, embedding_size), requires_grad=True)

        # Set value for out_dim_g
        # Set NNs sizes
        a = max(4 * concatenated_shape, embedding_size, 64)
        in_dim_g = embedding_size + concatenated_shape
        in_dim_f = embedding_size + out_dim_g
        self.l = _generate_fully_connected(
            input_dim=in_dim_g,
            output_dim=out_dim_g,
            hidden_dims=[
                a,
            ]
            * num_layers_g,
        )
        self.zeta = _generate_fully_connected(
            input_dim=in_dim_f,
            output_dim=concatenated_shape,
            hidden_dims=[
                a,
            ]
            * num_layers_zeta,
        )
        self.w = torch.nn.Parameter(torch.zeros((num_nodes, num_nodes)), requires_grad=True)

    def forward(self, samples: torch.Tensor, graphs: torch.Tensor) -> torch.Tensor:
        """
        Computes non-linear function hᵢ(X, G) using the given adjacency matrix.

            hᵢ(x, G) =  ζᵢ(Σⱼ Wᵢⱼ Gⱼᵢ lⱼ(xⱼ)

        We also use an embedding u so:

            hᵢ(x, G) =  ζ(uᵢ, Σⱼ Wᵢⱼ Gⱼᵢ l(uⱼ, xⱼ))

        l takes inputs of size batch_shape + (embedding_size + concatenated_shape) and outputs batch_shape + (out_dim_g)
        the input will be appropriately masked to correspond to one variable group

        ζ takes inputs of size batch_shape + (embedding_size + out_dim_g) and outputs batch_shape + (concatenated_shape)
        the ouptut is then masked to correspond to one variable

        Args:
            samples: tensor of shape batch_shape_x + batch_shape_f + batch_shape_g + [n_cols]
            graph: tensor of shape batch_shape_g + [n_nodes, n_nodes]
        Returns:
            tensor of shape batch_shape_x + batch_shape_f + batch_shape_g + [n_cols]
        """
        batch_shape_samples = samples.shape[:-1]
        batch_shape_g = graphs.shape[:-2]
        if len(batch_shape_g) > 0 and batch_shape_samples[-len(batch_shape_g) :] != batch_shape_g:
            raise ValueError(
                f"Batch shape of samples and graph must match but got {batch_shape_samples} and {batch_shape_g}"
            )

        # Shape batch_shape_x + batch_shape_f + batch_shape_g + (num_nodes, concatenated_shape)
        masked_samples = torch.einsum("...i,ji->...ji", samples, self.group_mask)
        # Shape batch_shape_x + batch_shape_f + batch_shape_g + (num_nodes, embedding_size)
        expanded_embed = self.embeddings.expand(*batch_shape_samples, -1, -1)

        # l(uⱼ, xⱼ) Shape batch_shape_x + batch_shape_f + batch_shape_g + (num_nodes, embedding_size + concatenated_shape)
        encoded_samples = self.l(
            torch.cat([masked_samples, expanded_embed], dim=-1)  # (concatenate xⱼ and embeddings uⱼ)
        )  # Shape batch_shape_x + batch_shape_f + batch_shape_g + (num_nodes, out_dim_g)

        # Aggregate sum and generate input for f (concatenate X_aggr and embeddings)
        # Σⱼ Wᵢⱼ Gⱼᵢ l(uⱼ, xⱼ) Shape batch_shape_x + batch_shape_f + batch_shape_g + (num_nodes, out_dim_g)
        encoded_samples_aggr = torch.einsum("...jk,...jl->...lk", encoded_samples, self.w * graphs)

        # ζ(uᵢ, Σⱼ Wᵢⱼ Gⱼᵢ l(uⱼ, xⱼ)) Shape batch_shape_x + batch_shape_f + batch_shape_g + (num_nodes, concatenated_shape)
        decoded_samples = self.zeta(torch.cat([encoded_samples_aggr, expanded_embed], dim=-1))

        # Mask and aggregate Shape batch_shape_x + batch_shape_f + batch_shape_g + (concatenated_shape)
        return torch.einsum("...ij,ij->...j", decoded_samples, self.group_mask)


def _generate_fully_connected(
    input_dim: int,
    output_dim: int,
    hidden_dims: list[int],
) -> nn.Module:
    """
    Generate a fully connected network.

    Args:
        input_dim: Int. Size of input to network.
        output_dim: Int. Size of output of network.
        hidden_dims: List of int. Sizes of internal hidden layers. i.e. [a, b] is three linear layers with shapes (input_dim, a), (a, b), (b, output_dim)

    Returns:
        Sequential object containing the desired network.
    """
    layers: list[nn.Module] = []

    prev_dim = input_dim
    for idx, hidden_dim in enumerate(hidden_dims):

        block: list[nn.Module] = []

        if idx > 0:
            block.append(nn.LayerNorm(prev_dim))
        block.extend([nn.Linear(prev_dim, hidden_dim), nn.LeakyReLU()])

        seq_block: nn.Module = nn.Sequential(*block)
        if prev_dim == hidden_dim:
            seq_block = _ResBlock(seq_block)
        layers.append(seq_block)

        prev_dim = hidden_dim

    layers.extend([nn.LayerNorm(prev_dim), nn.Linear(prev_dim, output_dim)])

    return nn.Sequential(*layers)


class _ResBlock(nn.Module):
    """
    Wraps an nn.Module, adding a skip connection to it.
    """

    def __init__(self, block: nn.Module):
        """
        Args:
            block: module to which skip connection will be added. The input dimension must match the output dimension.
        """
        super().__init__()
        self.block = block

    def forward(self, x):
        return x + self.block(x)



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Optional

import mlflow
import pytorch_lightning as pl
import torch
from pytorch_lightning.cli import LightningArgumentParser, Namespace, SaveConfigCallback
from pytorch_lightning.trainer.states import TrainerFn
from pytorch_lightning.utilities.types import STEP_OUTPUT

from causica.training.auglag import AugLagLossCalculator, AugLagLR


class AuglagLRCallback(pl.Callback):
    """Wrapper Class to make the Auglag Learning Rate Scheduler compatible with Pytorch Lightning"""

    def __init__(self, scheduler: AugLagLR, log_auglag: bool = False, disabled_epochs: Optional[set[int]] = None):
        """
        Args:
            scheduler: The auglag learning rate scheduler to wrap.
            log_auglag: Whether to log the auglag state as metrics at the end of each epoch.
        """
        self.scheduler = scheduler
        self._log_auglag = log_auglag
        self._disabled_epochs = disabled_epochs

    def on_train_batch_end(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule, outputs: STEP_OUTPUT, batch: Any, batch_idx: int
    ) -> None:
        _ = trainer
        _ = batch
        _ = batch_idx
        assert isinstance(outputs, dict)
        optimizer = pl_module.optimizers()
        assert isinstance(optimizer, torch.optim.Optimizer)
        auglag_loss: AugLagLossCalculator = pl_module.auglag_loss  # type: ignore

        # Disable if we reached a disabled epoch - disable, otherwise make sure the scheduler is enabled
        if self._disabled_epochs and trainer.current_epoch in self._disabled_epochs:
            self.scheduler.disable(auglag_loss)
        else:
            self.scheduler.enable(auglag_loss)

        is_converged, convergence_reasons = self.scheduler.step(
            optimizer=optimizer,
            loss=auglag_loss,
            loss_value=outputs["loss"],
            lagrangian_penalty=outputs["constraint"],
        )

        # Notify trainer to stop if the auglag algorithm has converged
        if is_converged:
            pl_module.log(
                "auglag_inner_convergence_reason",
                convergence_reasons.inner_convergence_reason.value,
                on_epoch=True,
                rank_zero_only=True,
                prog_bar=False,
            )
            pl_module.log(
                "auglag_outer_convergence_reason",
                convergence_reasons.outer_convergence_reason.value,
                on_epoch=True,
                rank_zero_only=True,
                prog_bar=False,
            )
            trainer.should_stop = True

    def on_train_epoch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        _ = trainer
        if self._log_auglag:
            auglag_state = {
                "num_lr_updates": self.scheduler.num_lr_updates,
                "outer_opt_counter": self.scheduler.outer_opt_counter,
                "step_counter": self.scheduler.step_counter,
                "outer_below_penalty_tol": self.scheduler.outer_below_penalty_tol,
                "outer_max_rho": self.scheduler.outer_max_rho,
                "last_best_step": self.scheduler.last_best_step,
                "last_lr_update_step": self.scheduler.last_lr_update_step,
            }
            pl_module.log_dict(auglag_state, on_epoch=True, rank_zero_only=True, prog_bar=False)


class MLFlowSaveConfigCallback(SaveConfigCallback):
    """Logs the config using MLFlow if there is an active run, otherwise saves locally as the superclass."""

    def __init__(
        self,
        parser: LightningArgumentParser,
        config: Namespace,
        config_filename: str = "config.yaml",
        multifile: bool = False,
    ) -> None:
        super().__init__(
            parser=parser, config=config, config_filename=config_filename, overwrite=True, multifile=multifile
        )

    def setup(self, trainer: pl.Trainer, pl_module: pl.LightningModule, stage: TrainerFn) -> None:  # type: ignore
        # Save the file on rank 0
        if trainer.is_global_zero and stage == TrainerFn.FITTING:
            with TemporaryDirectory() as tmpdir:
                temporary_config_path = str(Path(tmpdir) / self.config_filename)
                self.parser.save(
                    self.config,
                    temporary_config_path,
                    skip_none=False,
                    overwrite=self.overwrite,
                    multifile=self.multifile,
                )
                # AzureML throws a raw Exception if the artifact already exists, so we check the error message
                try:
                    mlflow.log_artifact(temporary_config_path)
                # pylint: disable=broad-exception-caught
                except Exception as e:
                    if "Resource Conflict" not in str(e.args):
                        raise e



from typing import Optional

import torch
import torch.distributions as td
from torch import nn

from causica.distributions.noise.noise import IndependentNoise, Noise, NoiseModule


class UnivariateNormalNoise(td.Normal, Noise[torch.Tensor]):
    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return samples - self.loc

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return noise + self.loc


class UnivariateNormalNoiseRescaled(UnivariateNormalNoise):
    """
    This is a `UnivariateNormalNoise` where the scale parameter is used in sample/noise transformations.
    """

    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return (samples - self.loc) / self.scale

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return self.scale * noise + self.loc


class UnivariateNormalNoiseModule(NoiseModule[IndependentNoise[UnivariateNormalNoise]]):
    """Represents a UnivariateNormalNoise with learnable parameters for independent variables."""

    def __init__(self, dim: int, init_log_scale: float = 0.0):
        """
        Args:
            dim: Number of dimensions for the NormalNoise.
        """
        super().__init__()
        self.log_scale = nn.Parameter(torch.full(torch.Size([dim]), init_log_scale))

    def forward(
        self, x: Optional[tuple[torch.Tensor, torch.Tensor] | torch.Tensor] = None
    ) -> IndependentNoise[UnivariateNormalNoise]:

        """
        Generarate Independent noise module for the Gaussian distribution.

        If both loc and log_scale are provided, use them to generate the noise module.
        If only loc is provided, use the loc and the learnable log_scale.
        If neither loc nor log_scale is provided, use zeros as loc and the learnable log_scale.

        Args:
            x: Tuple of loc and log_scale or just the loc.
        Returns:
            Independent noise module.
        """

        if isinstance(x, tuple):
            x, y = x
            return IndependentNoise(UnivariateNormalNoiseRescaled(loc=x, scale=torch.exp(y)), 1)

        if isinstance(x, torch.Tensor):
            y = self.log_scale
        else:
            x = torch.zeros_like(self.log_scale)
            y = self.log_scale
        return IndependentNoise(UnivariateNormalNoise(loc=x, scale=torch.exp(y)), 1)



from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution
from causica.distributions.adjacency.constrained_adjacency_distributions import (
    ConstrainedAdjacency,
    ConstrainedAdjacencyDistribution,
    LaggedConstrainedAdjacencyDistribution,
    TemporalConstrainedAdjacency,
    TemporalConstrainedAdjacencyDistribution,
)
from causica.distributions.adjacency.edges_per_node_erdos_renyi import EdgesPerNodeErdosRenyiDAGDistribution
from causica.distributions.adjacency.enco import ENCOAdjacencyDistribution, ENCOAdjacencyDistributionModule
from causica.distributions.adjacency.erdos_renyi import ErdosRenyiDAGDistribution
from causica.distributions.adjacency.fixed_adjacency_distribution import FixedAdjacencyDistribution
from causica.distributions.adjacency.geometric_random_graph import GeometricRandomGraphDAGDistribution
from causica.distributions.adjacency.gibbs_dag_prior import ExpertGraphContainer, GibbsDAGPrior
from causica.distributions.adjacency.scale_free import ScaleFreeDAGDistribution
from causica.distributions.adjacency.stochastic_block_model import StochasticBlockModelDAGDistribution
from causica.distributions.adjacency.temporal_adjacency_distributions import (
    LaggedAdjacencyDistribution,
    RhinoLaggedAdjacencyDistribution,
    RhinoLaggedAdjacencyDistributionModule,
    TemporalAdjacencyDistribution,
    TemporalAdjacencyDistributionModule,
)
from causica.distributions.adjacency.three_way import ThreeWayAdjacencyDistribution
from causica.distributions.adjacency.watts_strogatz import WattsStrogatzDAGDistribution



import torch

from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution


class FixedAdjacencyDistribution(AdjacencyDistribution):
    """Delta distribution over a fixed adjacency matrix."""

    def __init__(self, adjacency: torch.Tensor, validate_args: bool | None = None):
        if adjacency.shape[-1] != adjacency.shape[-2]:
            raise ValueError("Adjacency matrix must be square")
        if adjacency.ndim != 2:
            raise ValueError("Adjacency matrix must be 2 dimensional")

        self.adjacency = adjacency
        self.num_nodes = adjacency.shape[-1]
        super().__init__(num_nodes=self.num_nodes, validate_args=validate_args)

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        """
        Return the underlying adjacency matrix.

        Args:
            sample_shape: the shape of the samples to return
        Returns:
            A tensor of shape sample_shape + batch_shape + (num_nodes, num_nodes)
        """
        _ = temperature
        return self.sample(sample_shape=sample_shape)

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Return the underlying adjacency matrix.

        Args:
            sample_shape: the shape of the samples to return
        Returns:
            A tensor of shape sample_shape + batch_shape + (num_nodes, num_nodes)
        """
        return self.adjacency.expand(sample_shape + self.adjacency.shape)

    def entropy(self) -> torch.Tensor:

        return torch.zeros([])

    @property
    def mean(self) -> torch.Tensor:
        return self.adjacency

    @property
    def mode(self) -> torch.Tensor:
        return self.adjacency

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError



import inspect
from collections import defaultdict
from typing import Callable

import pytorch_lightning as pl
from torch.utils.data import ConcatDataset, DataLoader

from causica.data_generation.samplers.sem_sampler import SEMSampler
from causica.datasets.samplers import SubsetBatchSampler
from causica.datasets.synthetic_dataset import CausalMetaset


class SyntheticDataModule(pl.LightningDataModule):
    """A datamodule to produce datasets and their underlying causal graphs and interventions.

    This datamodule samples synthetic datasets from a list of SEM samplers and returns batches of CausalDataset objects.
    This means, that the underlying tensors are not yet stacked and will need to be processed by the module using the
    data. Currently, the data module uses the `SubsetBatchSampler` to sample batches from the different sampled SEMs,
    ensuring that each batch is sampled from SEMs with the same number of nodes. This is important, as it allows for
    easy batching of the data, as the tensors are already of the same shape.

    Note:
        This data module does currently not support DDP, because the batch sampler does not simply consume indices
        from a a regular sampler. In future, DDP support will be added by ensuring that each compute node is individually
        seeded to generate different data samples and adapting the batch sampler to ensure that an epoch has the same length
        regardless of the number of compute nodes.

        This data module might generate duplicate data samples, when num_workers > 1. This is because the individual
        workers are not individually seeded.
    """

    def __init__(
        self,
        sem_samplers: list[SEMSampler] | Callable[[], list[SEMSampler]],
        train_batch_size: int,
        test_batch_size: int,
        dataset_size: int,
        num_interventions: int = 0,
        num_intervention_samples: int = 100,
        num_sems: int = 0,
        batches_per_metaset: int = 1,
        sample_interventions: bool = False,
        sample_counterfactuals: bool = False,
        treatment_variable: str | None = None,
        effect_variables: list[str] | None = None,
        num_workers: int = 0,
        pin_memory: bool = True,
        persistent_workers: bool = True,
        prefetch_factor: int = 16,
    ) -> None:
        """
        Args:
            sem_samplers: Either a list of sem samplers or a function that returns a list of SEM samplers
            train_batch_size: The training batch size to use
            test_batch_size: The testing batch size to use
            dataset_size: The size of dataset to generate
            num_interventions: The number of interventions to generate (0 for no interventions)
            num_sems: The number of SEMs to generate (0 for infinite SEMs)
            batches_per_metaset: The number of batches per epoch per dataset
            sample_interventions: Whether to sample interventions
            sample_counterfactuals: Whether to sample counterfactuals
            treatment_variable: This specify the name of the nodes to be taken as treatment from the generated sems.
                The sem sampler must always generate this treatment.
            effet_variables: This specify the names of the nodes to be taken as effects from the generated sems.
                The sem sampler must always generate this effect.
            num_workers: The number of workers to use for the dataloader
            pin_memory: Whether to pin memory for the dataloader
            persistent_workers: Whether to use persistent workers for the dataloader
            prefetch_factor: The prefetch factor to use for the dataloader
        """
        super().__init__()
        self.dataset_size = dataset_size
        self.train_batch_size = train_batch_size
        self.test_batch_size = test_batch_size
        if callable(sem_samplers) and inspect.isclass(type(sem_samplers)):
            self.sem_samplers_hparams = vars(sem_samplers)
        elif callable(sem_samplers):
            self.sem_samplers_hparams = {k: v.default for k, v in inspect.signature(sem_samplers).parameters.items()}
        else:
            self.sem_samplers_hparams = {}

        self.sem_sampler_list = sem_samplers if isinstance(sem_samplers, list) else sem_samplers()
        self.num_interventions = num_interventions
        self.num_intervention_samples = num_intervention_samples
        self.num_workers = num_workers
        self.num_sems = num_sems
        self.batches_per_metaset = batches_per_metaset
        self.sample_interventions = sample_interventions
        self.sample_counterfactuals = sample_counterfactuals
        self.treatment_variable = treatment_variable
        self.effect_variables = effect_variables

        self.num_workers = num_workers
        self.persistent_workers = persistent_workers
        self.pin_memory = pin_memory
        self.prefetch_factor = prefetch_factor

        self.save_hyperparameters(logger=False)
        self.train_dataset = self._get_dataset(self.train_batch_size)
        self.val_dataset = self._get_dataset(self.test_batch_size)

    def prepare_data(self) -> None:
        # Log hyperparameters if trainer is available
        if self.trainer and self.trainer.logger:
            hyperparams = {k: v for k, v in self.hparams.items() if k != "sem_samplers"}
            hyperparams["sem_samplers_hparams"] = self.sem_samplers_hparams

            self.trainer.logger.log_hyperparams(hyperparams)

    def _get_dataset(self, dataset_size: int) -> ConcatDataset:
        """Builds causal datasets given the SEM samplers.

        Args:
            dataset_size: Number of samples of the causal dataset (ie number of datasets generated).

        Returns:
            dataset object
        """
        cur_dataset_size = dataset_size * self.batches_per_metaset
        if self.num_sems > 0 and cur_dataset_size < self.num_sems:
            raise ValueError(
                f"Dataset size must be at least the number of SEMs. Got {cur_dataset_size} < {self.num_sems}"
            )
        datasets_by_nodes = defaultdict(list)
        for sampler in self.sem_sampler_list:
            datasets_by_nodes[sampler.adjacency_dist.num_nodes].append(
                CausalMetaset(
                    sampler,
                    sample_dataset_size=self.dataset_size,
                    dataset_size=cur_dataset_size,
                    num_interventions=self.num_interventions,
                    num_intervention_samples=self.num_intervention_samples,
                    num_sems=self.num_sems,
                    sample_interventions=self.sample_interventions,
                    sample_counterfactuals=self.sample_counterfactuals,
                    treatment_variable=self.treatment_variable,
                    effect_variables=self.effect_variables,
                )
            )

        return ConcatDataset(ConcatDataset(ds) for ds in datasets_by_nodes.values())

    def train_dataloader(self):
        batch_sampler = SubsetBatchSampler([len(d) for d in self.train_dataset.datasets], self.train_batch_size)
        return DataLoader(
            dataset=self.train_dataset,
            batch_sampler=batch_sampler,
            collate_fn=lambda x: x,
            num_workers=self.num_workers,
            persistent_workers=self.persistent_workers,
            pin_memory=self.pin_memory,
            prefetch_factor=self.prefetch_factor,
        )

    def val_dataloader(self):
        batch_sampler = SubsetBatchSampler([len(d) for d in self.val_dataset.datasets], self.test_batch_size)
        return DataLoader(
            dataset=self.val_dataset, batch_sampler=batch_sampler, collate_fn=lambda x: x, pin_memory=self.pin_memory
        )

    def test_dataloader(self):
        return self.val_dataloader()



import torch
import torch.distributions as td
from torch.distributions.constraints import Constraint

from causica.distributions.adjacency import AdjacencyDistribution
from causica.distributions.distribution_module import DistributionModule
from causica.distributions.noise.joint import JointNoiseModule
from causica.functional_relationships import FunctionalRelationships, TemporalEmbedFunctionalRelationships
from causica.sem.distribution_parameters_sem import DistributionParametersSEM
from causica.sem.temporal_distribution_parameters_sem import TemporalDistributionParametersSEM


class SEMDistribution(td.Distribution):
    """A distribution over structural equation models.

    Samples are instances of DistributionParametersSEM. Note however that this was created before
    pytorch set the expected type of samples to torch.Tensor, so this is breaking the types a bit.

    The distribution is essentially the same as the given adjacency distribution but with samples converted to SEMs.
    Therefore, all distribution properties such as entropy, mean and mode are given by the equivalent properties for the
    adjacency distribution.
    """

    arg_constraints: dict[str, Constraint] = {}

    def __init__(
        self,
        adjacency_dist: AdjacencyDistribution,
        noise_module: JointNoiseModule,
        functional_relationships: FunctionalRelationships,
    ):
        """
        Args:
            adjacency_dist: Distribution from which adjacency matrices are sampled to construct SEMs.
            noise_module: The noise module for any SEM of this distribution.
            functional_relationships: The functional relationship for any SEM of this distribution.
        """
        super().__init__()
        self._adjacency_dist = adjacency_dist
        self._noise_module = noise_module
        self._functional_relationships = functional_relationships

    def _create_sems(self, graphs: torch.Tensor) -> list[DistributionParametersSEM]:
        graphs = graphs.reshape(-1, *self._adjacency_dist.event_shape)
        return [
            DistributionParametersSEM(
                graph=graph,
                noise_dist=self._noise_module,
                func=self._functional_relationships,
            )
            for graph in graphs.unbind(dim=0)
        ]

    def sample(self, sample_shape: torch.Size = torch.Size()):
        graphs = self._adjacency_dist.sample(sample_shape)
        if not sample_shape:
            graphs = graphs[None, ...]
        return self._create_sems(graphs)

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0):
        graphs = self._adjacency_dist.relaxed_sample(sample_shape=sample_shape, temperature=temperature)
        return self._create_sems(graphs)

    def entropy(self) -> torch.Tensor:
        return self._adjacency_dist.entropy()

    @property
    def mean(self) -> DistributionParametersSEM:  # type: ignore
        return self._create_sems(self._adjacency_dist.mean)[0]

    @property
    def mode(self) -> DistributionParametersSEM:  # type: ignore
        return self._create_sems(self._adjacency_dist.mode)[0]

    def log_prob(self, value: DistributionParametersSEM) -> torch.Tensor:  # type: ignore
        return self._adjacency_dist.log_prob(value.graph)


class SEMDistributionModule(DistributionModule[SEMDistribution]):
    """Represents a SEMDistribution with learnable parameters."""

    def __init__(
        self,
        adjacency_module: DistributionModule[AdjacencyDistribution],
        functional_relationships: FunctionalRelationships,
        noise_module: JointNoiseModule,
    ):
        super().__init__()
        self.adjacency_module = adjacency_module
        self.functional_relationships = functional_relationships
        self.noise_module = noise_module

    def forward(self) -> SEMDistribution:
        return SEMDistribution(
            adjacency_dist=self.adjacency_module(),
            noise_module=self.noise_module,
            functional_relationships=self.functional_relationships,
        )


class TemporalSEMDistribution(SEMDistribution):
    def __init__(
        self,
        adjacency_dist: AdjacencyDistribution,
        noise_module: JointNoiseModule,
        functional_relationships: TemporalEmbedFunctionalRelationships,
    ):
        super().__init__(adjacency_dist, noise_module, functional_relationships)
        self._functional_relationships: TemporalEmbedFunctionalRelationships = functional_relationships

    def _create_sems(self, graphs: torch.Tensor) -> list[TemporalDistributionParametersSEM]:  # type: ignore
        graphs = graphs.reshape(-1, *self._adjacency_dist.event_shape)
        return [
            TemporalDistributionParametersSEM(
                graph=graph,
                noise_dist=self._noise_module,
                func=self._functional_relationships,
            )
            for graph in graphs.unbind(dim=0)
        ]


class TemporalSEMDistributionModule(DistributionModule[TemporalSEMDistribution]):
    """Represents a SEMDistribution with learnable parameters."""

    def __init__(
        self,
        adjacency_module: DistributionModule[AdjacencyDistribution],
        functional_relationships: TemporalEmbedFunctionalRelationships,
        noise_module: JointNoiseModule,
    ):
        super().__init__()
        self.adjacency_module = adjacency_module
        self.functional_relationships = functional_relationships
        self.noise_module = noise_module

    def forward(self) -> TemporalSEMDistribution:
        return TemporalSEMDistribution(
            adjacency_dist=self.adjacency_module(),
            noise_module=self.noise_module,
            functional_relationships=self.functional_relationships,
        )



import abc

import torch
from tensordict import TensorDict

from causica.distributions.transforms import TensorToTensorDictTransform


class FunctionalRelationships(abc.ABC, torch.nn.Module):
    def __init__(self, shapes: dict[str, torch.Size], batch_shape: torch.Size = torch.Size([])) -> None:
        """Base class for functional relationships.

        Args:
            shapes: Dict of node shapes (how many dimensions a node has)
                Order corresponds to the order in graph(s).
        """
        super().__init__()
        self.batch_shape = batch_shape
        self.shapes = shapes
        # create a transform for mapping tensors to tensordicts
        self.tensor_to_td = TensorToTensorDictTransform(shapes)
        # this needs to be registered to the module, and register buffer doesn't work
        self.stacked_key_masks = torch.nn.Parameter(self.tensor_to_td.stacked_key_masks(), requires_grad=False)

    @abc.abstractmethod
    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        """Calculates the predictions of the children from parents.

        Functional relationships expect samples to have a batch shape in order: samples_shape, functions_shape,
        graphs_shape. The graphs are expected have a matching batch shape graphs_shape. This then applies the functional
        relationship to each sample using the corresponding function and graph. This allows for batched processing of
        different functions (e.g. interventions) or graphs.

        Args:
            samples: dictionary of variable samples of shape batch_size_x + batch_size_f + batch_shape_g + [node shape].
            graphs: tensor of shape batch_size_g + [nodes, nodes]

        Returns:
            Dictionary of torch.Tensors of shape batch_size_x + batch_size_f + batch_shape_g + [node shape]
        """



import torch
from torch import nn

from causica.nn.deci_embed_nn import _generate_fully_connected


class TemporalEmbedNN(nn.Module):
    """Defines functional relationships for variables in a timeseries.

    Follows the model described in:
        Gong, W., Jennings, J., Zhang, C. and Pawlowski, N., 2022.
        Rhino: Deep causal temporal relationship learning with history-dependent noise.
        ICLR 2023
        arXiv:2210.14706

    The model is a generalization of the DECI model to include temporal dependencies but represents relationships
    between the current timestep and a fixed number of steps backwards.

    For each variable xᵢ we use
        ζᵢ(x) = ζ(eᵢ,  Σ_{k in pa(i)} l(eₖ, xₖ)),

    where eᵢ is a learned embedding for node i and pa(i) includes the variables from any timestep in the context.
    """

    def __init__(
        self,
        group_mask: torch.Tensor,
        embedding_size: int,
        out_dim_l: int,
        num_layers_l: int,
        num_layers_zeta: int,
        context_length: int,
    ):
        """
        Args:
            group_mask: A mask of shape (num_nodes, num_processed_cols) such that group_mask[i, j] = 1. when col j is in
                group i.
            embedding_size: Size of the embeddings used by each node. Also affects the hidden dimensionality of l and ζ,
                which is set to the larger of 4 * concatenated_shape, embedding_size and 64.
            out_dim_l: Output dimension of the "inner" NN, l. If none, default is embedding size.
            num_layers_l: Number of layers in the "inner" NN, l.
            num_layers_zeta: Number of layers in the "outer" NN, ζ.
            context_length: The total number of timesteps in the context of the model, including the current timestep.
        """
        super().__init__()
        self.group_mask = group_mask
        num_nodes, concatenated_shape = group_mask.shape
        self.context_length = context_length

        # Initialize embeddings uₜⱼ and graph weights Wₜᵢⱼ
        # Notice that we store the embeddings and the graph with the time axis in the reversed order from the paper,
        # e.g. embedding[lag - t] are the embeddings for the nodes t steps behind, where t=0 corresponds to the
        # instantaneous effect. lag, in the paper, is equivalent to context_length - 1, and is the number of steps back
        # from the current that's visible to the model.
        self.embeddings = nn.parameter.Parameter(0.01 * torch.randn(context_length, num_nodes, embedding_size))
        self.w = torch.nn.Parameter(torch.zeros((context_length, num_nodes, num_nodes)), requires_grad=True)

        # Set NNs sizes
        a = max(4 * concatenated_shape, embedding_size, 64)
        in_dim_g = embedding_size + concatenated_shape
        in_dim_f = embedding_size + out_dim_l
        self.l = _generate_fully_connected(
            input_dim=in_dim_g,
            output_dim=out_dim_l,
            hidden_dims=[a] * num_layers_l,
        )
        self.zeta = _generate_fully_connected(
            input_dim=in_dim_f,
            output_dim=concatenated_shape,
            hidden_dims=[a] * num_layers_zeta,
        )

    def forward(self, samples: torch.Tensor, graphs: torch.Tensor) -> torch.Tensor:
        """Forward of the module.

        Computes non-linear function hᵢ(X, G) using the given adjacency matrix.

            hᵢ(x, G) =  ζᵢ(Σₜⱼ Wₜⱼᵢ Gₜⱼᵢ lⱼ(xₜⱼ)

        We also use an embedding u so:

            hᵢ(x, G) =  ζ(uₗᵢ, Σₜⱼ Wₜⱼᵢ Gₜⱼᵢ l(xₜⱼ, uₜⱼ))

        l takes inputs of size batch_shape + (context_length, embedding_size + concatenated_shape) and outputs
        batch_shape + (out_dim_g). The input will be appropriately masked to correspond to one variable group.

        ζ takes inputs of size batch_shape + (embedding_size + out_dim_g, ) and outputs batch_shape +
        (concatenated_shape, ) the ouptut is then masked to correspond to one variable.

        Args:
            samples: tensor of shape batch_shape_x + batch_shape_f + batch_shape_g + [context_length, n_cols]
            graph: tensor of shape batch_shape_g + [context_length, n_nodes, n_nodes]

        Returns:
            tensor of shape batch_shape_x + batch_shape_f + batch_shape_g + [n_cols]
        """
        batch_shape_samples = samples.shape[:-2]
        batch_shape_g = graphs.shape[:-3]
        if len(batch_shape_g) > 0 and batch_shape_samples[-len(batch_shape_g) :] != batch_shape_g:
            raise ValueError(
                f"Batch shape of samples and graph must match but got {batch_shape_samples} and {batch_shape_g}"
            )

        # Shape batch_shape_x + batch_shape_f + batch_shape_g + (context_length, num_nodes, concatenated_shape)
        masked_samples = torch.einsum("...i,ji->...ji", samples, self.group_mask)
        # Shape batch_shape_x + batch_shape_f + batch_shape_g + (context_length, num_nodes, embedding_size)
        expanded_embed = self.embeddings.expand(*batch_shape_samples, -1, -1, -1)

        # Shape batch_shape_x + batch_shape_f + batch_shape_g +
        #     (context_length, num_nodes, concatenated_shape + embedding_size)
        x_and_embeddings = torch.cat([masked_samples, expanded_embed], dim=-1)  # (concatenate xⱼ and embeddings uⱼ)
        # l(uⱼ, xⱼ): Shape batch_shape_x + batch_shape_f + batch_shape_g + (context_length, num_nodes, out_dim_g)
        encoded_samples = self.l(x_and_embeddings)

        # Shape batch_shape_samples + batch_shape_g + (num_nodes, out_dim_g)
        aggregated_effects = torch.einsum("...lij,...lio->...jo", self.w * graphs, encoded_samples)

        # ζ(uᵢ, Σⱼ Wᵢⱼ Gⱼᵢ l(uⱼ, xⱼ))
        # Shape batch_shape_x + batch_shape_f + batch_shape_g + (num_nodes, concatenated_shape)
        decoded_samples = self.zeta(torch.cat([aggregated_effects, expanded_embed[..., -1, :, :]], dim=-1))

        # Mask and aggregate Shape batch_shape_x + batch_shape_f + batch_shape_g + (concatenated_shape)
        return torch.einsum("...ij,ij->...j", decoded_samples, self.group_mask)



import torch
from tensordict import TensorDict, TensorDictBase

from causica.datasets.tensordict_utils import expand_td_with_batch_shape
from causica.distributions import JointNoiseModule
from causica.functional_relationships import DoFunctionalRelationships, TemporalEmbedFunctionalRelationships
from causica.sem.structural_equation_model import SEM


def split_lagged_and_instanteneous_values(td: TensorDictBase) -> tuple[TensorDict, TensorDict]:
    """Splits a temporal TensorDict into lagged and instantaneous values.

    Args:
        td: The temporal TensorDict with shape (..., context_length, (variable_shape)). The batch size of the TensorDict
            should not include the context length, ie. td.batch_size = (...) and not td.batch_size = (..., context_length).

    Returns:
        A tuple with the lagged values and the instantaneous values.
    """
    orig_batch_size = td.batch_size
    new_batch_size = next(td.values()).shape[:-1]
    td.batch_size = new_batch_size

    lagged, instantaneous = td[..., :-1], td[..., -1]

    td.batch_size = orig_batch_size
    lagged.batch_size = orig_batch_size
    instantaneous.batch_size = orig_batch_size

    return lagged, instantaneous


def concatenate_lagged_and_instaneous_values(
    lagged_values: TensorDictBase,
    instantaneous_values: TensorDictBase,
) -> TensorDict:
    """Concatenate the lagged and instantaneous values.

    Args:
        instantaneous_values: The TensorDict with the instantaneous values with shape (..., (variable_shape))
        lagged_values: The TensorDict with the lagged values with shape (..., context_length, (variable_shape))

    Returns:
        A new TensorDict with the concatenated lagged and instantaneous values.
    """
    return TensorDict(
        {
            k: torch.cat([lagged_values[k], v[..., None, :]], dim=-2)
            for k, v in instantaneous_values.items(include_nested=True, leaves_only=True)
        },
        batch_size=instantaneous_values.batch_size,
    )


class TemporalDistributionParametersSEM(SEM):
    """
    A Structural Equation Model where the functional forward pass generates the parameters of the distributions to be sampled.

    This is more general than the simple additive case.
    """

    def __init__(
        self,
        graph: torch.Tensor,
        noise_dist: JointNoiseModule,
        func: TemporalEmbedFunctionalRelationships,
    ):
        """
        Args:
            graph: Adjacency Matrix
            noise_dist: A dictionary of callables for generating distributions from predictions
            func: The functional relationship that can map from the tensor representation to the to itself.
        """

        self.batch_shape_f = func.batch_shape
        batch_shape_g = torch.Size(graph.shape[:-3])
        self.context_length = graph.shape[-3]

        if self.context_length != func.nn.context_length:
            raise ValueError(
                f"Graph context length {self.context_length} does not match functional relationship context length {func.context_length}"
            )

        if len(self.batch_shape_f) == 0 and len(batch_shape_g) > 0:
            self.batch_shape_f = torch.Size((1,))
            func.batch_shape = self.batch_shape_f

        super().__init__(
            graph=graph,
            node_names=noise_dist.keys(),
            batch_shape=torch.Size(self.batch_shape_f + batch_shape_g),
        )
        self.noise_dist = noise_dist
        self.func = func

    def log_prob(self, value: TensorDict) -> torch.Tensor:
        """
        Compute the log prob of the observations

        Args:
            value: a dictionary of `sample_shape + batch_shape + context_length + [variable_dim]` shaped tensors
        Return:
            A tensor of shape `sample_shape + batch_shape` of the log probability of each event
        """
        if value[value.sorted_keys[0]].shape[-2] != self.context_length:
            raise ValueError(
                f"Value shape {value[value.sorted_keys[0]].shape} does not match graph shape {self.graph.shape}."
                " The last batch dimension of the value should match the context length."
            )
        expanded_value = expand_td_with_batch_shape(value, self.batch_shape)
        _lagged, test_value = split_lagged_and_instanteneous_values(expanded_value)  # pylint: disable=unused-variable
        return self.noise_dist(self.func(expanded_value, self.graph)).log_prob(test_value)

    def noise_to_sample(self, noise: TensorDict) -> TensorDict:
        """
        For a given noise vector, return the corresponding sample values.

        Args:
            noise: Dictionary of tensors representing the noise shape
            `sample_shape + batch_shape + context_length + noise_shape`, where the context length dimension contains
            the observed history as well as the noise for the current time step.
        Return:
            Dictionary of samples from the sem corresponding to the noise, shape
            `sample_shape + batch_shape + event_shape`
        """
        x = expand_td_with_batch_shape(noise.clone(), self.batch_shape).zero_()
        history, noise = split_lagged_and_instanteneous_values(noise)
        expanded_noise = expand_td_with_batch_shape(noise, self.batch_shape)
        for _ in self.node_names:
            new_sample = self.noise_dist(self.func(x, self.graph)).noise_to_sample(expanded_noise)
            x.update_(concatenate_lagged_and_instaneous_values(history, new_sample))
        x = split_lagged_and_instanteneous_values(x)[-1]
        if isinstance(self.func, DoFunctionalRelationships):
            do = self.func.do
            if do.batch_dims > 0:
                do = do[(...,) + (None,) * len(self.batch_shape_g)]
            expanded_do = do.expand(*x.batch_size)
            x.update(expanded_do, inplace=True)
            x = self.func.func.tensor_to_td.order_td(x)
        return x

    def sample_to_noise(self, sample: TensorDict) -> TensorDict:
        """
        For a given sample get the noise vector.

        Args:
            sample: samples from the sem each of shape `sample_shape + batch_shape + context_length + event_shape`
        Return:
            Dictionary of tensors representing the noise shape `sample_shape + batch_shape + noise_shape`
        """
        if sample[sample.sorted_keys[0]].shape[-2] != self.context_length:
            raise ValueError(
                f"Value shape {sample[sample.sorted_keys[0]].shape} does not match graph shape {self.graph.shape}."
                " The last batch dimension of the value should match the context length."
            )

        expanded_sample = expand_td_with_batch_shape(sample, self.batch_shape)
        _lagged, test_sample = split_lagged_and_instanteneous_values(expanded_sample)  # pylint: disable=unused-variable
        return self.noise_dist(self.func(expanded_sample, self.graph)).sample_to_noise(test_sample)

    def do(self, interventions: TensorDict) -> "TemporalDistributionParametersSEM":
        """Return the SEM associated with the interventions.

        If the interventions have a batch shape, then the returned functions will have the same batch shape to batch
        the computation across interventions.
        """
        _ = interventions
        raise NotImplementedError("Interventions are not yet supported for temporal graphs")

    @torch.no_grad()
    def sample_noise(self, sample_shape: torch.Size = torch.Size()) -> TensorDict:
        """
        Sample the noise vector for the distribution.

        Args:
            sample_shape: shape of the returned noise samples
        Return:
            Dictionary of samples from the noise distribution for each node, shape
            `sample_shape + batch_shape + event_shape`.
        """

        return self.noise_dist().sample(torch.Size(torch.Size(sample_shape) + self.batch_shape))



import math

import igraph as ig
import numpy as np
import torch

from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution


class WattsStrogatzDAGDistribution(AdjacencyDistribution):
    """
    An adjacency distribution for sampling Directed Acyclic Graphs using Watts Strogatz.

    Re-implementation in pytorch from AVICI:

    https://github.com/larslorch/avici

    """

    def __init__(
        self,
        num_nodes: int,
        lattice_dim: list[int],
        rewire_prob: list[float],
        neighbors: list[int],
        validate_args: bool | None = None,
    ):
        """
        Args:
            num_nodes: the number of nodes in the DAGs to be sampled
            lattice_dim: List of the dimension of the lattice
            rewire_prob: List of the rewiring probability
            neighbors: List of the number of neighbors used
            validate_args: Optional arguments from AdjacencyDistribution
        """
        assert num_nodes > 0, "Number of nodes in the graph must be greater than 0"
        super().__init__(num_nodes=num_nodes, validate_args=validate_args)
        self.lattice_dim = lattice_dim
        self.rewire_prob = rewire_prob
        self.neighbors = neighbors

    def _one_sample(self):
        """
        Sample a binary adjacency matrix from the underlying distribution.

        Returns:
            A tensor of shape (num_nodes, num_nodes)
        """
        get_lattice_dim = np.random.choice(self.lattice_dim).item()
        # choose size s.t. we get at smallest possible n_vars greater than requested n_vars given the dimension of lattice
        dim_size = math.ceil(self.num_nodes ** (1.0 / get_lattice_dim))
        get_rewire_prob = np.random.choice(self.rewire_prob).item()
        get_neighbors = np.random.choice(self.neighbors).item()
        graph = ig.Graph.Watts_Strogatz(
            dim=get_lattice_dim, size=dim_size, nei=get_neighbors, p=get_rewire_prob, multiple=False, loops=False
        )

        # drop excessive vertices s.t. we get exactly n_vars
        n_excessive = len(graph.vs) - self.num_nodes
        assert n_excessive >= 0
        if n_excessive:
            graph.delete_vertices(np.random.choice(graph.vs, size=n_excessive, replace=False))
        assert (
            len(graph.vs) == self.num_nodes
        ), f"Didn't get requested graph; g.vs: {len(graph.vs)}, n_vars {self.num_nodes}"

        # make directed
        dag = np.triu(np.array(graph.get_adjacency().data).astype(int), k=1)

        # randomly permute
        perm = np.random.permutation(self.num_nodes).tolist()
        dag = dag[perm, :][:, perm]

        if not ig.Graph.Weighted_Adjacency(dag.tolist()).is_dag():
            raise ValueError("Sampled graph is not a DAG")

        dag = torch.tensor(dag, dtype=torch.float32)
        return dag

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        raise NotImplementedError

    def sample(self, sample_shape: torch.Size = torch.Size()):
        """
        Sample binary adjacency matrices from the underlying distribution.

        Args:
            sample_shape: The shape of the samples to be drawn.
        """
        num_samples = int(torch.prod(torch.tensor(sample_shape)).item())
        return torch.stack([self._one_sample() for _ in range(num_samples)]).reshape(
            sample_shape + (self.num_nodes, self.num_nodes)
        )

    def entropy(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mean(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mode(self) -> torch.Tensor:
        raise NotImplementedError

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError



from typing import Optional

import torch
import torch.distributions as td
from torch import nn

from causica.distributions.noise.noise import IndependentNoise, Noise, NoiseModule


class UnivariateCauchyNoise(td.Cauchy, Noise[torch.Tensor]):
    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return samples - self.loc

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return noise + self.loc


class UnivariateCauchyNoiseRescaled(UnivariateCauchyNoise):
    """
    This is a `UnivariateCauchyeNoise` where the scale parameter is used in sample/noise transformations.
    """

    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return (samples - self.loc) / self.scale

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return self.scale * noise + self.loc


class UnivariateCauchyNoiseModule(NoiseModule[IndependentNoise[UnivariateCauchyNoise]]):
    """Represents a UnivariateCauchyNoise with learnable parameters for independent variables."""

    def __init__(self, dim: int, init_log_scale: float | torch.Tensor = 0.0):
        """
        Args:
            dim: Number of dimensions for the NormalNoise.
        """
        super().__init__()
        if isinstance(init_log_scale, torch.Tensor):
            if init_log_scale.squeeze().ndim == 0:
                init_log_scale = torch.full(torch.Size([dim]), init_log_scale.item())
            else:
                assert init_log_scale.ndim == 1
                assert init_log_scale.shape[0] == dim
        else:
            init_log_scale = torch.full(torch.Size([dim]), init_log_scale)

        self.log_scale = nn.Parameter(init_log_scale)

    def forward(
        self, x: Optional[tuple[torch.Tensor, torch.Tensor] | torch.Tensor] = None
    ) -> IndependentNoise[UnivariateCauchyNoise]:
        """
        Generarate Independent noise module for the Cauchy distribution.

        If both loc and log_scale are provided, use them to generate the noise module.
        If only loc is provided, use the loc and the learnable log_scale.
        If neither loc nor log_scale is provided, use zeros as loc and the learnable log_scale.

        Args:
            x: Tuple of loc and log_scale or just the loc.
        Returns:
            Independent noise module.
        """

        if isinstance(x, tuple):
            assert len(x) == 2, "Expected a tuple of length 2."
            x, y = x
            return IndependentNoise(UnivariateCauchyNoiseRescaled(loc=x, scale=torch.exp(y)), 1)

        if isinstance(x, torch.Tensor):
            y = self.log_scale
        else:
            x = torch.zeros_like(self.log_scale)
            y = self.log_scale
        return IndependentNoise(UnivariateCauchyNoise(loc=x, scale=torch.exp(y)), 1)



import torch
from torch.utils.data import Dataset

from causica.data_generation.generate_data import sample_dataset
from causica.data_generation.samplers.sem_sampler import SEMSampler
from causica.datasets.causal_dataset import CausalDataset
from causica.distributions.transforms import TensorToTensorDictTransform


class CausalMetaset(Dataset):
    """A metaset (dataset of datasets) that returns datasets from SEM samples.

    SEM samples are returned as a CausalDataset object.

    See Also:
        CausalDataset: For a description of the format of the samples.
    """

    def __init__(
        self,
        sem_sampler: SEMSampler,
        sample_dataset_size: int,
        dataset_size: int,
        num_interventions: int = 0,
        num_intervention_samples: int = 1000,
        num_sems: int = 0,
        sample_interventions: bool = False,
        sample_counterfactuals: bool = False,
        treatment_variable: str | None = None,
        effect_variables: list[str] | None = None,
    ):
        """
        Args:
            sem_sampler: The sampler for SEMs
            sample_dataset_size: The size of the dataset to sample from the SEM
            dataset_size: The size of this dataset, if not supplied it will be infinitely long
                It is useful to set this value to a finite size so it can be used with `ChainDataset`,
                which relies on the iterator terminating to chain the next one.
            num_interventions: The number of interventions to sample per dataset. If 0, no interventions are sampled.
            num_intervention_samples: The number of interventional samples to sample.
            num_sems: The number of sems to sample the data from. If 0, each data sample is generated from a new SEM.
            sample_interventions: Whether to sample interventions.
            sample_counterfactuals: Whether to sample counterfactuals.
            treatment_variable: This specify the name of the nodes to be taken as treatment from the generated sems.
                The sem sampler must always generate this treatment.
            effet_variables: This specify the names of the nodes to be taken as effects from the generated sems.
                The sem sampler must always generate this effect.
        """
        self.sem_sampler = sem_sampler
        self.sample_dataset_size = torch.Size([sample_dataset_size])
        self.dataset_size = dataset_size
        self.num_interventions = num_interventions
        self.num_intervention_samples = num_intervention_samples
        self.num_sems = num_sems
        self.sample_interventions = sample_interventions
        self.sample_counterfactuals = sample_counterfactuals
        self.treatment_variable = treatment_variable
        self.effect_variables = effect_variables

        self.sems = [sem_sampler.sample() for _ in range(num_sems)]
        self.td_to_tensor_transform = TensorToTensorDictTransform(self.sem_sampler.shapes_dict)

    def __len__(self) -> int:
        """Return the size of the dataset."""
        return self.dataset_size

    def __getitem__(self, index) -> CausalDataset:
        """Return a sample from the dataset.

        See Also:
            CausalDataset: For a description of the format of the samples.
        """
        if index >= len(self) or index < -len(self):
            raise IndexError(f"index {index} out of range for dataset of size {len(self)}")
        return self._sample(index)

    def _sample(self, index: int = 0) -> CausalDataset:
        """Sample a new dataset and returns the data, graph and potentially interventions and counterfactuals."""
        if self.num_sems > 0:
            sem = self.sems[index % self.num_sems]
        else:
            sem = self.sem_sampler.sample()

        return sample_dataset(
            sem,
            self.sample_dataset_size,
            self.num_interventions,
            self.num_intervention_samples,
            self.sample_interventions,
            self.sample_counterfactuals,
            self.treatment_variable,
            self.effect_variables,
        )



from typing import Optional

import torch
import torch.distributions as td


class ExpertGraphContainer(torch.nn.Module):
    def __init__(self, dag: torch.Tensor, mask: torch.Tensor, confidence: float, scale: float) -> None:
        """Container holding an "experts" prior belief about the underlying causal DAG.

        Arguments:
            dag: The binary adjacency matrix representing domain knowledge in the form of a DAG. Corresponds to `mask`.
            mask: A binary mask indicating whether or not the corresponding edge of the `dag` has information.
            confidence: A value in the interval (0, 1] indicating the confidence of the existence of the edge
            scale: Scaling factor for expert graph loss
        """
        super().__init__()

        self.dag = torch.nn.Parameter(dag, requires_grad=False)
        self.mask = torch.nn.Parameter(mask, requires_grad=False)
        self.confidence: torch.Tensor
        self.scale: torch.Tensor
        self.register_buffer("confidence", torch.tensor(confidence, dtype=torch.float))
        self.register_buffer("scale", torch.tensor(scale, dtype=torch.float))


class GibbsDAGPrior(td.Distribution):
    """
    Represents a prior distribution over adjacency matrices.

    The prior distribution consists of two terms:
        A sparsity term
        A Expert Graph term that represents some known prior belief about the graph.

    Each term has an associated parameter (lambda)

    The event shape of the prior is either (num_nodes, num_nodes) or (context_length, num_nodes, num_nodes) for temporal
    models and is set at initialization. The shape is checked for given matrices.
    """

    arg_constraints: dict = {}

    def __init__(
        self,
        num_nodes: int,
        sparsity_lambda: float,
        expert_graph_container: Optional[ExpertGraphContainer] = None,
        context_length: int | None = None,
        **kwargs,
    ) -> None:
        """
        Args:
            num_nodes: Number of nodes in the graph
            sparsity_lambda: Coefficient of sparsity term
            dagness_alpha: Coefficient of dagness term
            dagness_rho: Coefficient of squared dagness term
            expert_graph_container: Dataclass containing prior belief about the real graph
            context_length: Optional, specifying the context_length in temporal models
        """

        event_shape = (num_nodes, num_nodes) if context_length is None else (context_length, num_nodes, num_nodes)
        super().__init__(torch.Size(), event_shape=torch.Size(event_shape), **kwargs)

        if expert_graph_container is not None:
            if expert_graph_container.dag.shape != event_shape:
                raise ValueError(
                    f"Expert graph shape {expert_graph_container.dag.shape} does not match event shape {event_shape}"
                )
            if expert_graph_container.mask.shape != event_shape:
                raise ValueError(
                    f"Expert graph mask shape {expert_graph_container.mask.shape} does not match event shape {event_shape}"
                )

        self._num_nodes = num_nodes
        self._expert_graph_container = expert_graph_container
        self._sparsity_lambda = sparsity_lambda
        self._context_length = context_length

    def get_sparsity_term(self, A: torch.Tensor) -> torch.Tensor:
        """
        A term that encourages sparsity (see https://arxiv.org/pdf/2106.07635.pdf).
        The term is small when A is sparse.

        Args:
            A (torch.Tensor): Adjacency matrix of shape event_shape.

        Returns:
            Sparsity term.
        """
        return A.abs().sum()

    def get_expert_graph_term(self, A: torch.Tensor) -> torch.Tensor:
        """
        A term that encourages A to be close to given expert graph.

        Args:
            A (torch.Tensor): Adjacency matrix of shape event_shape.

        Returns:
            (torch.Tensor): Expert graph term.

        """
        assert isinstance(self._expert_graph_container, ExpertGraphContainer)
        assert A.shape[-len(self.event_shape) :] == self.event_shape
        return (
            (
                self._expert_graph_container.mask
                * (A - self._expert_graph_container.confidence * self._expert_graph_container.dag)
            )
            .abs()
            .sum()
        )

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Calculates the (un-normalized) log probability of adjacency matrix `value`
        under the distribution given by this instance.

        Args:
            value (torch.Tensor): Adjacency matrix of shape event_shape.

        Returns:
            (torch.Tensor): The un-normalized log probability of `value`.

        """
        assert value.shape[-len(self.event_shape) :] == self.event_shape

        log_prob = -self._sparsity_lambda * self.get_sparsity_term(value)

        if self._expert_graph_container is not None:
            log_prob -= self._expert_graph_container.scale * self.get_expert_graph_term(value)

        return log_prob



import torch
from torch import nn


class PositionalEncodingFixed(nn.Module):
    """
    Standard positional encoding.

    """

    def __init__(self, d_model: int, max_seq_length: int):
        super().__init__()

        pe = torch.zeros(max_seq_length, d_model)
        position = torch.arange(0, max_seq_length, dtype=torch.float).unsqueeze(1)
        div_term = torch.pow(1e5, torch.linspace(0.0, 1.0, d_model // 2 + 1))

        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)

        self.pe: torch.Tensor
        self.register_buffer("pe", pe.unsqueeze(0))

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        """
        Args:
            inp: expected shape: (batch_size, max_seq_length, d_model)

        Returns:
            out: expected shape: (batch_size, max_seq_length, d_model)

        """

        return inp + self.pe[: inp.size(1), :]


class Encoding(nn.Module):
    """Learnable encoding.

    Either for nodes or for positions.
    For the nodes, we multiply (in the hadamard sense) the node value with the positional embedding

    """

    def __init__(self, d_model: int, max_seq_length: int, encoding_type: str = "node"):
        super().__init__()
        self.emb = nn.Embedding(max_seq_length, d_model)

        self.position: torch.Tensor
        self.register_buffer("position", torch.arange(0, max_seq_length, dtype=torch.long))

        self.encoding_type = encoding_type

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        """
        Args:
            inp: expected shape: (batch_size, max_seq_length, 1)

        Returns:
            out: expected shape: (batch_size, max_seq_length, d_model)

        """
        if self.encoding_type == "node":
            return inp * self.emb(self.position)

        if self.encoding_type == "position":
            return inp + self.emb(self.position)

        raise ValueError("Unknown encoding type")


class PositionWiseFeedForward(nn.Module):
    def __init__(self, d_model: int, d_ff: int, add_prod: bool = False):
        super().__init__()

        self.fc1 = nn.Linear(d_model, d_ff)
        self.fc2 = nn.Linear(d_ff, d_model)
        self.relu = nn.ReLU()
        self.add_prod = add_prod
        if add_prod:
            self.fc_m = nn.Linear(d_model, d_ff)

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        """
        Args:
            inp: expected shape: (batch_size, max_seq_length, d_model)

        Returns:
            out: expected shape: (batch_size, max_seq_length, d_model)

        """
        if self.add_prod:
            return self.fc2(self.relu(self.fc1(inp)) * self.fc_m(inp))

        return self.fc2(self.relu(self.fc1(inp)))



from causica.functional_relationships.deci_functional_relationships import DECIEmbedFunctionalRelationships
from causica.functional_relationships.do_functional_relationships import (
    DoFunctionalRelationships,
    create_do_functional_relationship,
)
from causica.functional_relationships.functional_relationships import FunctionalRelationships
from causica.functional_relationships.heteroscedastic_rff_functional_relationships import (
    HeteroscedasticRFFFunctionalRelationships,
)
from causica.functional_relationships.linear_functional_relationships import LinearFunctionalRelationships
from causica.functional_relationships.rff_functional_relationships import RFFFunctionalRelationships
from causica.functional_relationships.temporal_functional_relationships import TemporalEmbedFunctionalRelationships



from typing import Generic, Mapping, TypeVar

import torch
import torch.distributions as td
from tensordict import TensorDictBase
from torch import nn

from causica.distributions.transforms.base import TransformModule, TypedTransform


class JointTransform(TypedTransform[TensorDictBase, TensorDictBase]):
    """A joint transform that applies a different transform to each key in the TensorDict.

    Keys in the input that are not found in the transform are left unchanged.

    This is heavily inspired by the `torch.distributions.transforms.StackTransform` class.
    See https://pytorch.org/docs/stable/distributions.html#torch.distributions.transforms.StackTransform
    """

    def __init__(self, transformations: Mapping[str, td.Transform], cache_size: int = 0):
        """
        Args:
            transformations: A dictionary of transforms, where the keys are the keys in the TensorDict
            cache_size: Size of cache. If zero, no caching is done. If one, the latest single value is cached.
                Only 0 and 1 are supported.
        """
        bad_transformation_types = {type(t) for t in transformations.values() if not isinstance(t, td.Transform)}
        if bad_transformation_types:
            raise TypeError(
                "All transformations must be subtypes of `torch.distributions.Transform`, but the "
                f"following are not: {bad_transformation_types} are not."
            )
        if cache_size:
            transformations = {key: t.with_cache(cache_size) for key, t in transformations.items()}
        super().__init__(cache_size=cache_size)
        self.transformations = transformations

    def _call(self, x: TensorDictBase) -> TensorDictBase:
        return x.clone().update(
            {key: transform(x[key]) for key, transform in self.transformations.items() if key in x.keys()}
        )

    def _inverse(self, y: TensorDictBase) -> TensorDictBase:
        # We cannot use ._inv as pylint complains with E202: _inv is hidden because of `self._inv = None`
        # in td.Transform.__init__

        return y.clone().update(
            {key: transform.inv(y[key]) for key, transform in self.transformations.items() if key in y.keys()}
        )

    def log_abs_det_jacobian(self, x: TensorDictBase, y: TensorDictBase) -> TensorDictBase:
        if set(x.keys()) != set(y.keys()):
            raise ValueError("x and y must have the same keys.")

        if not set(self.transformations.keys()) <= set(x.keys()):
            raise ValueError("All keys in transformations must be in x and y.")

        return x.clone().update(
            {
                key: self.transformations[key].log_abs_det_jacobian(x[key], y[key])
                if key in self.transformations
                else torch.zeros_like(x[key])
                for key in x.keys()
            }
        )

    @property
    def bijective(self):
        return all(t.bijective for t in self.transformations.values())

    @property
    def domain(self):
        return {key: t.domain for key, t in self.transformations.items()}

    @property
    def codomain(self):
        return {key: t.codomain for key, t in self.transformations.items()}


T_co = TypeVar("T_co", bound=nn.Module, covariant=True)


class _TypedModuleDict(Generic[T_co], nn.ModuleDict, Mapping[str, T_co]):
    """Allow a ModuleDict to be interpreted as a mapping."""

    def __hash__(self) -> int:
        return nn.ModuleDict.__hash__(self)


class JointTransformModule(JointTransform, TransformModule[TensorDictBase, TensorDictBase]):
    """Joint transform with TransformModule transformations applied per key to a TensorDict."""

    def __init__(self, transformations: Mapping[str, TransformModule], *args, **kwargs):
        """
        Args:
            transformations: A mapping of transforms, where the keys are the keys in the TensorDict.
            *args, **kwargs: Passed to the JointTransform.
        """
        super().__init__(transformations, *args, **kwargs)
        self.transformations = _TypedModuleDict[TransformModule](transformations)



import torch
from tensordict import TensorDict, TensorDictBase
from torch.distributions.constraints import Constraint

from causica.distributions.transforms.base import TypedTransform


class TensorToTensorDictTransform(TypedTransform[torch.Tensor, TensorDictBase]):
    """
    A transform for converting a torch tensor to a TensorDict.

    It extracts the slices from the last dimension of the tensor and assigns them to the correct key.
    """

    arg_constraints: dict[str, Constraint] = {}
    bijective = True

    def __init__(self, shapes: dict[str, torch.Size]):
        """
        Args:
            shapes: the shapes of each of the keys
        """
        super().__init__()
        self.shapes = shapes
        self.num_keys = len(shapes)
        self.output_shape, self.slices = shapes_to_slices(self.shapes)

    def _call(self, x: torch.Tensor) -> TensorDictBase:
        """Create a Tensordict by retrieving the slice associated with each key."""
        return TensorDict({name: x[..., slice_] for name, slice_ in self.slices.items()}, batch_size=x.shape[:-1])

    def _inverse(self, y: TensorDictBase) -> torch.Tensor:
        """
        Create a tensor by stacking the slice associated with each key.

        Args:
            y: Tensordict with batch_shape
        Returns:
            A tensor with shape batch_shape + [output_shape]
        """
        return torch.cat([y[name] for name in self.slices], dim=-1)

    def log_abs_det_jacobian(self, _: torch.Tensor, y: TensorDictBase) -> TensorDictBase:
        """This transformation doesn't affect the log det jacobian"""
        return y.apply(torch.zeros_like)

    def stacked_key_masks(self) -> torch.Tensor:
        """
        Create a binary of matrix of where each key is in the tensor.

        Returns:
            A matrix of shape [num_keys, output_shape] with 1 if the index of the tensor
            belongs to the key corresponding to that row
        """
        stacked_key_masks = torch.zeros((self.num_keys, self.output_shape), dtype=torch.float)
        for i, slice_ in enumerate(self.slices.values()):
            stacked_key_masks[i, slice_] = 1.0
        return stacked_key_masks

    def order_td(self, td: TensorDictBase) -> TensorDictBase:
        """Order the keys of a TensorDict to match the order of the shapes."""
        return td.select(*self.shapes.keys(), inplace=True)


def shapes_to_slices(shapes: dict[str, torch.Size]) -> tuple[int, dict[str, slice]]:
    """
    Convert a dictionary of shapes to a dictionary of masks by stacking the shapes

    Each mask corresponds to the embedded location in the tensor

    Args:
        shapes: A dict of key names to shapes

    Returns:
        The shape of the stacked tensor and a dictionary of each key to the mask
    """
    assert all(len(shape) == 1 for shape in shapes.values())

    slices: dict[str, slice] = {}
    idx = 0
    next_idx = 0
    for name, shape in shapes.items():
        next_idx = idx + shape[-1]
        slices[name] = slice(idx, next_idx)
        idx = next_idx

    return next_idx, slices



import torch
from torch import nn

from causica.nn.fip.decoder_layer import AmortizedDecoderLayer


class AmortizedEncoder(nn.Module):
    """
    Causal Dataset Encoder Model.

    Takes a batch of datasets and encodes it using multiple `AmortizedDecoderLayer`s.

    This is a reinterpretation of the AVICI model in PyTorch:

    https://github.com/larslorch/avici

    """

    def __init__(
        self,
        d_model: int,
        num_heads: int,
        dim_key: int,
        num_layers: int,
        d_ff: int,
        dropout: float,
    ):
        super().__init__()

        self.emb_data = nn.Linear(1, d_model)
        self.decoder_layers_nodes = nn.ModuleList(
            [
                AmortizedDecoderLayer(
                    d_model,
                    num_heads,
                    dim_key,
                    d_ff,
                    dropout,
                    max_seq_length=2,
                    attn_type="standard",
                    cost_type="dot_product",
                )
                for _ in range(num_layers)
            ]
        )

        self.decoder_layers_samples = nn.ModuleList(
            [
                AmortizedDecoderLayer(
                    d_model,
                    num_heads,
                    dim_key,
                    d_ff,
                    dropout,
                    max_seq_length=2,
                    attn_type="standard",
                    cost_type="dot_product",
                )
                for _ in range(num_layers)
            ]
        )

        self.norm = nn.LayerNorm(d_model)
        self.num_layers = num_layers

    def forward(self, tgt: torch.Tensor) -> torch.Tensor:
        """
        Args:
            tgt: expected shape (*, num_samples, max_seq_length) or (*, num_samples, max_seq_length, 2)

        Returns:
            output: expected shape: (*, num_samples, max_seq_length, d_model)

        """
        dec_output = tgt.unsqueeze(-1)

        # embed the data
        dec_output = self.emb_data(dec_output)

        for k in range(self.num_layers):
            dec_output = self.decoder_layers_nodes[k](dec_output, dec_output, dec_output, None)

            dec_output = dec_output.transpose(-2, -3)
            dec_output = self.decoder_layers_samples[k](dec_output, dec_output, dec_output, None)
            dec_output = dec_output.transpose(-2, -3)

        dec_output = self.norm(dec_output)

        return dec_output



from typing import Any, Iterable, Optional, Type, Union

from torch import Tensor
from torchmetrics import Metric, MetricCollection
from torchmetrics.regression import MeanAbsoluteError, MeanAbsolutePercentageError
from torchmetrics.wrappers import MultitaskWrapper


class MeanAbsoluteErrorWithThreshold(MeanAbsoluteError):
    """This compute the MAE with a minimum filter on the target values.

    This compute the MAE for target that is higher than the min_threshold.
    """

    def __init__(self, min_threshold: float, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.threshold = min_threshold

    def update(self, preds: Tensor, target: Tensor) -> None:
        """Update state with filtered predictions and targets."""
        # filter based on the threshold
        mask = target > self.threshold
        filtered_preds = preds[mask]
        filtered_target = target[mask]
        super().update(filtered_preds, filtered_target)


class MeanAbsolutePercentageErrorWithThreshold(MeanAbsolutePercentageError):
    """This compute the MAPE with a minimum filter on the target values.

    This compute the MAPE for target that is higher than the min_threshold.
    """

    def __init__(self, min_threshold: float, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.threshold = min_threshold

    def update(self, preds: Tensor, target: Tensor) -> None:
        """Update state with filtered predictions and targets."""
        # filter based on the threshold
        mask = target > self.threshold
        filtered_preds = preds[mask]
        filtered_target = target[mask]
        super().update(filtered_preds, filtered_target)


def create_metrics_for_variables(
    variables: Iterable[str],
    metrics: MetricCollection,
    min_thresholds: Optional[dict[str, float]] = None,
    threshold_metrics: Optional[dict[str, Type[Metric]]] = None,
) -> MultitaskWrapper:
    """
    This function will create a MultiTaskWrapper obj with keys to be the variable names.

    Args:
        variables: Variables name list to calculate the metrics for.
        metrics: MetricCollection to store the metrics we want to compute.
        min_thresholds: A dictionary containing the threshold values for each of the variables.
        threshold_metrics: The dict containing the thresholded torchmetrics class. If None, then no additional threshold
            metrics will be added.

    Returns:
        MultitaskWrapper obj with keys to be the variable names.
    """

    metrics_dict: dict[str, Union[Metric, MetricCollection]] = {
        key: metrics.clone(postfix=f".{key}") for key in variables
    }

    if threshold_metrics is not None and min_thresholds is not None:
        for variable_name, threshold in min_thresholds.items():
            for name, cur_threshold_metric_class in threshold_metrics.items():
                cur_threshold_metric = cur_threshold_metric_class(min_threshold=threshold)
                metrics_dict[variable_name].add_module(name, cur_threshold_metric)

    return MultitaskWrapper(metrics_dict)


def filter_metrics_wrapper(variable_list: Optional[list[str]], metrics_wrapper: MultitaskWrapper) -> MultitaskWrapper:
    """
    This will filter the MultiTaskWrapper obj to select only the variables in variable_list. If None, the metrics will not be filtered.

    Args:
        variable_list: List of variables to filter the metrics_dict.
        metrics_wrapper: MultitaskWrapper obj with keys to be the variable names.

    Returns:
        Filtered MultitaskWrapper obj.
    """
    if variable_list is None:
        return metrics_wrapper

    filtered_dict = {key: metrics_wrapper.task_metrics[key] for key in variable_list}

    return MultitaskWrapper(filtered_dict)



import dataclasses
import json
import os
from typing import Any, Callable, Iterable, Optional

import fsspec
import numpy as np
import torch
from tensordict import TensorDict

from causica.datasets.causica_dataset_format.load import (
    CounterfactualWithEffects,
    DataEnum,
    InterventionWithEffects,
    Variable,
    VariablesMetadata,
    get_group_names,
)
from causica.datasets.interventional_data import CounterfactualData, InterventionData
from causica.datasets.variable_types import VariableTypeEnum


class TensorEncoder(json.JSONEncoder):
    """A JSON encoder that can encode tensors, numpy arrays and TensorDicts."""

    def default(self, o):
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, torch.Tensor):
            return self.default(o.numpy())
        if isinstance(o, TensorDict):
            return o.to_dict()
        if isinstance(o, set):
            return list(o)
        return json.JSONEncoder.default(self, o)


class VariableEncoder(json.JSONEncoder):
    """A JSON encoder that can encode Variable objects."""

    def default(self, o):
        if isinstance(o, VariablesMetadata):
            return dataclasses.asdict(o)
        if isinstance(o, Variable):
            return dataclasses.asdict(o)
        if isinstance(o, VariableTypeEnum):
            return o.value
        return json.JSONEncoder.default(self, o)


def _get_idx_list_from_node_names(names: Iterable[str], node_names: Iterable[str]) -> list[int]:
    """Returns a list of indices corresponding to the node names in the node_names list.

    Args:
        names: A list of node names
        node_names: A list of all node names

    Returns:
        A list of indices corresponding to the node names in the node_names list
    """
    return [np.argwhere(np.array(node_names) == name).item() for name in names]


def _packed_concatenate_tensordict_values(
    td: TensorDict, node_names: Iterable[str], variable_types: dict[str, VariableTypeEnum]
) -> np.ndarray:
    """Concatenates the values of a TensorDict along the last dimension. The order of the values is determined by the node_names list.

    Args:
        td: A TensorDict
        node_names: The names of the nodes in the TensorDict

    Returns:
        A numpy array of the concatenated values
    """

    def _maybe_pack_onehot(t: torch.Tensor, variable_type: VariableTypeEnum):
        return torch.argmax(t, dim=-1, keepdim=True) if variable_type == VariableTypeEnum.CATEGORICAL else t

    return np.concatenate(
        [_maybe_pack_onehot(td[name], variable_types[name]) for name in node_names if name in td.keys()], axis=-1
    )


def intervention_data_to_dict(
    intervention: InterventionData,
    reference: InterventionData | None,
    variables_metadata: VariablesMetadata,
    effect_idx_names: Iterable[str] | None = None,
) -> dict[str, np.ndarray | None]:
    """Converts an InterventionData object to a dictionary that can be saved to a json file.

    Args:
        intervention: An InterventionData object
        reference: An optional InterventionData object to use as a reference
        variables_metadata: The variables metadata for the dataset
        effect_idx_names: The names of the variables to use as the effect

    Returns:
        A dictionary of a single intervention data sample in causica dataset format that can be saved to a json file
    """
    node_names = get_group_names(variables_metadata)
    variable_types = {var.group_name: var.type for var in variables_metadata.variables}

    effect_idxs = _get_idx_list_from_node_names(effect_idx_names or [], node_names)
    intervention_idxs = _get_idx_list_from_node_names(intervention.intervention_values.keys(), node_names)

    intervention_values = _packed_concatenate_tensordict_values(
        intervention.intervention_values, node_names, variable_types
    )
    test_data = _packed_concatenate_tensordict_values(intervention.intervention_data, node_names, variable_types)

    reference_data = None
    intervention_reference = None
    if reference is not None:
        reference_data = _packed_concatenate_tensordict_values(reference.intervention_data, node_names, variable_types)
        intervention_reference = _packed_concatenate_tensordict_values(
            reference.intervention_values, node_names, variable_types
        )

    return {
        "intervention_idxs": np.array(intervention_idxs),
        "intervention_values": intervention_values,
        "test_data": test_data,
        "reference_data": reference_data,
        "intervention_reference": intervention_reference,
        "effect_idxs": np.array(effect_idxs),
        "conditioning_idxs": None,
        "conditioning_values": None,
    }


def counterfactual_data_to_dict(
    counterfactual: CounterfactualData,
    reference: Optional[CounterfactualData],
    variables_metadata: VariablesMetadata,
    effect_idx_names: Iterable[str] | None = None,
) -> dict[str, np.ndarray | list[int] | None]:
    """Converts a CounterfactualData object to a dictionary that can be saved to a json file.

    Args:
        counterfactual: A CounterfactualData object
        reference: An optional CounterfactualData object to use as a reference
        variables_metadata: The variables metadata for the dataset
        effect_idx_names: The names of the variables to use as the effect

    Returns:
        A dictionary of a single counterfactual data sample in causica dataset format that can be saved to a json file
    """
    node_names = get_group_names(variables_metadata)
    variable_types = {var.group_name: var.type for var in variables_metadata.variables}

    effect_idxs = _get_idx_list_from_node_names(effect_idx_names or [], node_names)
    conditioning_idxs = _get_idx_list_from_node_names(counterfactual.factual_data.keys(), node_names)
    intervention_idxs = _get_idx_list_from_node_names(counterfactual.intervention_values.keys(), node_names)

    conditioning_values = _packed_concatenate_tensordict_values(counterfactual.factual_data, node_names, variable_types)
    intervention_values = _packed_concatenate_tensordict_values(
        counterfactual.intervention_values, node_names, variable_types
    )
    test_data = _packed_concatenate_tensordict_values(counterfactual.counterfactual_data, node_names, variable_types)

    reference_data = None
    intervention_reference = None
    if reference is not None:
        reference_data = _packed_concatenate_tensordict_values(
            reference.counterfactual_data, node_names, variable_types
        )
        intervention_reference = _packed_concatenate_tensordict_values(
            reference.intervention_values, node_names, variable_types
        )

    return {
        "intervention_idxs": intervention_idxs,
        "intervention_values": intervention_values,
        "test_data": test_data,
        "effect_idxs": effect_idxs,
        "conditioning_idxs": conditioning_idxs,
        "conditioning_values": conditioning_values,
        "reference_data": reference_data,
        "intervention_reference": intervention_reference,
    }


def intervention_or_counterfactual_to_causica_dict(
    data_with_effects: list[CounterfactualWithEffects] | list[InterventionWithEffects],
    variables_metadata: VariablesMetadata,
) -> dict[str, Any]:
    """Converts a list of CounterfactualWithEffects objects to a dictionary that can be saved to a json file.

    Args:
        data_with_effects: A list of CounterfactualWithEffects or InterventionWithEffects objects
        variables_metadata: The variables metadata for the dataset

    Returns:
        A dictionary of the data in causica dataset format that can be saved to a json file
    """
    node_names = get_group_names(variables_metadata)

    convert_fn: Callable
    if isinstance(data_with_effects[0][0], InterventionData):
        assert all(
            isinstance(data, InterventionData) and (reference is None or isinstance(reference, InterventionData))
            for data, reference, _ in data_with_effects
        )
        convert_fn = intervention_data_to_dict
        data_td = data_with_effects[0][0].intervention_data
    elif isinstance(data_with_effects[0][0], CounterfactualData):
        assert all(
            isinstance(data, CounterfactualData) and (reference is None or isinstance(reference, CounterfactualData))
            for data, reference, _ in data_with_effects
        )
        convert_fn = counterfactual_data_to_dict
        data_td = data_with_effects[0][0].factual_data
    else:
        raise ValueError(
            f"data_with_effects must be a list of InterventionData or CounterfactualData objects, got {type(data_with_effects[0][0])}"
        )

    columns_to_nodes = [idx for idx, name in enumerate(node_names) for _ in range(data_td[name].shape[-1])]

    return {
        "metadata": {
            "columns_to_nodes": columns_to_nodes,
        },
        "environments": [
            convert_fn(data_obj, reference_obj, variables_metadata, effect_idxs)
            for data_obj, reference_obj, effect_idxs in data_with_effects
        ],
    }


def save_data(
    savedir: str,
    data: VariablesMetadata
    | torch.Tensor
    | TensorDict
    | list[InterventionWithEffects]
    | list[CounterfactualWithEffects],
    data_enum: DataEnum | None = None,
    variables: VariablesMetadata | None = None,
    **storage_kwargs,
) -> None:
    """
    Save data to disk in the Causica dataset format.

    Args:
        savedir (str): The directory to save the data to. This can be any fsspec compatible url.
        data:
            The data to save. Must be one of the following types:
            - VariablesMetadata: If saving variables metadata.
            - torch.Tensor: If saving the true adjacency matrix.
            - TensorDict: If saving train, validation, or test data.
            - list[InterventionWithEffects]: If saving interventions data.
            - list[CounterfactualWithEffects]: If saving counterfactuals data.
        data_enum: The type of data being saved. Defaults to None. Not required for variables or adjacency.
        variables: The variables metadata associated with the data. Defaults to None. Not required for variables or adjacency.
        **storage_kwargs: Additional keyword arguments to pass to the fsspec.open function.

    Raises:
        AssertionError: If the input data is of an unexpected type or if the data_enum and variables arguments are not consistent with the input data type.
    """
    if isinstance(data, VariablesMetadata):
        assert data_enum is None or data_enum == DataEnum.VARIABLES_JSON
        with fsspec.open(os.path.join(savedir, "variables.json"), mode="w", **storage_kwargs) as f:
            json.dump(dataclasses.asdict(data), f, indent=4, cls=VariableEncoder)
    elif isinstance(data, torch.Tensor):
        assert data_enum is None or data_enum == DataEnum.TRUE_ADJACENCY
        with fsspec.open(os.path.join(savedir, "adj_matrix.csv"), mode="w", **storage_kwargs) as f:
            np.savetxt(
                f,
                data.numpy(),
                delimiter=",",
                fmt="%i",
            )
    elif isinstance(data, TensorDict):
        assert data_enum in (DataEnum.TRAIN, DataEnum.VALIDATION, DataEnum.TEST)
        assert variables is not None
        node_names = get_group_names(variables)
        variable_types = {var.group_name: var.type for var in variables.variables}
        with fsspec.open(os.path.join(savedir, data_enum.value), mode="w", **storage_kwargs) as f:
            np.savetxt(
                f,
                _packed_concatenate_tensordict_values(data, node_names, variable_types),
                delimiter=",",
            )
    else:
        assert (
            data_enum == DataEnum.COUNTERFACTUALS
            and isinstance(data, list)
            and isinstance(data[0][0], CounterfactualData)
        ) or (
            data_enum == DataEnum.INTERVENTIONS and isinstance(data, list) and isinstance(data[0][0], InterventionData)
        )
        assert variables is not None
        with fsspec.open(os.path.join(savedir, data_enum.value), mode="w", **storage_kwargs) as f:
            json.dump(
                intervention_or_counterfactual_to_causica_dict(data, variables),
                f,
                indent=4,
                cls=TensorEncoder,
            )


def save_dataset(
    savedir: str,
    variables: VariablesMetadata,
    adjacency: torch.Tensor,
    train_data: TensorDict,
    test_data: TensorDict,
    val_data: TensorDict | None = None,
    interventions: list[InterventionWithEffects] | None = None,
    counterfactuals: list[CounterfactualWithEffects] | None = None,
    overwrite: bool = False,
    **storage_kwargs,
) -> None:
    """
    Saves a dataset to a directory in the Causica format.

    Args:
        savedir: The directory to save the dataset to. This can be any fsspec compatible url.
        variables: Metadata about the variables in the dataset.
        adjacency: The adjacency matrix of the causal graph.
        train_data: The training data.
        test_data: The test data.
        val_data: The validation data.
        interventions: A list of interventions to save.
        counterfactuals: A list of counterfactuals to save.
        overwrite: Whether to overwrite the directory if it already exists. Defaults to False.
        **storage_kwargs: Additional keyword arguments to pass to the storage backend.
    """
    fs, _ = fsspec.core.url_to_fs(savedir, **storage_kwargs)
    if fs.exists(savedir):
        print(f"Directory {savedir} already exists...")
        if not overwrite:
            print("Skipping...")
            return
    fs.makedirs(savedir, exist_ok=True)

    save_data(savedir, variables, DataEnum.VARIABLES_JSON, variables, **storage_kwargs)
    save_data(savedir, adjacency, DataEnum.TRUE_ADJACENCY, variables, **storage_kwargs)
    save_data(savedir, train_data, DataEnum.TRAIN, variables, **storage_kwargs)
    save_data(savedir, test_data, DataEnum.TEST, variables, **storage_kwargs)
    if val_data is not None:
        save_data(savedir, val_data, DataEnum.VALIDATION, variables, **storage_kwargs)
    if interventions:
        save_data(savedir, interventions, DataEnum.INTERVENTIONS, variables, **storage_kwargs)
    if counterfactuals:
        save_data(savedir, counterfactuals, DataEnum.COUNTERFACTUALS, variables, **storage_kwargs)
